# """
# Testing Copied from the SSD Repo:
# https://github.com/eugenevinitsky/sequential_social_dilemma_games/blob/master/tests/test_envs.py
# """
# import unittest
#
# import numpy as np
# from gymnasium.spaces import Dict
#
# from cogrid.CoGridEnv_env import CoGridEnv
# from cogrid.core.agent import Agent
# from cogrid.core.grid_utils import ascii_to_numpy
# from cogrid.constants import DEFAULT_COLORS, GridConstants, BASE_ACTION_MAP
# from cogrid.features import map_to_image
#
#
# BASE_MAP = {"grid": ["#######", "#     #", "#     #", "#     #", "#     #", "#     #", "#######"], "values": {}}
# BASE_MAP_2 = {"grid": ["######", "# S  #", "#    #", "#    #", "#   S#", "######"], "values": {}}
#
#
# class DummyAgent(Agent):
#     def interact(self, char):
#         return GridConstants.FreeSpace
#
#
# class DummyMapEnv(CoGridEnv):
#     def __init__(self, config, test_grid_data):
#         self.test_grid_data = test_grid_data
#         self.start_positions = config.get("start_positions")
#         super().__init__(config=config, grid_creation_fn=self.grid_fn)
#
#     def _setup_agents(self):
#         map_with_agents = self.map_with_agents
#         for i in range(self.config["num_agents"]):
#             agent_id = f"agent-{i}"
#             spawn_point = self.start_positions.pop(0) if self.start_positions else self.select_spawn_point()
#             agent = DummyAgent(
#                 agent_id=agent_id,
#                 start_position=spawn_point,
#                 env=map_with_agents,
#                 config=self.config,
#             )
#             self.env_agents[agent_id] = agent
#
#     def grid_fn(self, **kwargs):
#         return self.test_grid_data
#
#
# class TestMapEnv(unittest.TestCase):
#     def tearDown(self) -> None:
#         self.env = None
#
#     def _construct_map(self, map, num_agents, start_positions, view_size=2, obs=("image_view", "ascii_view")):
#         dummy_config = {
#             "name": "dummyenv",
#             "horizon": 1000,
#             "view_size": view_size,
#             "num_agents": num_agents,
#             "start_positions": start_positions,
#             "features": obs,
#         }
#         self.env = DummyMapEnv(dummy_config, map)
#         self.env.reset()
#
#     def change_agent_position(self, agent_id, new_pos):
#         agent = self.env.env_agents[agent_id]
#         self.env.remove_agents_from_color_map()
#         agent.pos = new_pos
#         self.env.add_agents_to_color_map()
#
#     def add_agent_to_env(self, agent_id, start_position):
#         self.env.env_agents[agent_id] = DummyAgent(agent_id, start_position, self.env.map_with_agents, self.env.config)
#         self.env.feature_generators[agent_id] = [
#             self.env._fetch_feature_generator(ob_name) for ob_name in self.env.config["features"]
#         ]
#         self.env.observation_space[agent_id] = Dict(
#             {feature.name: feature.space for feature in self.env.feature_generators[agent_id]}
#         )
#         self.env.add_agents_to_color_map()
#
#     def test_step(self):
#         """Check that the step method works at all for all possible actions"""
#         self._construct_map(BASE_MAP_2, 1, None, 2)
#         aid = self.env.agent_ids[0]
#         action_dim = self.env.action_space.n
#         for i in range(action_dim):
#             self.env.step({aid: i})
#
#     def test_walls(self):
#         """Check that the spawned map and base map have walls in the right place"""
#         pad = 2
#         self._construct_map(BASE_MAP, 0, None, pad)
#         base_grid_np = ascii_to_numpy(BASE_MAP["grid"])
#         np.testing.assert_array_equal(self.env.base_grid[0, :], base_grid_np[0, :])
#         np.testing.assert_array_equal(self.env.base_grid[-1, :], base_grid_np[-1, :])
#         np.testing.assert_array_equal(self.env.base_grid[:, 0], base_grid_np[0, :])
#         np.testing.assert_array_equal(self.env.base_grid[:, -1], base_grid_np[:, -1])
#
#         wall_img = np.full(
#             (self.env.world_map_color.shape[1], self.env.world_map_color.shape[2], 3),
#             fill_value=DEFAULT_COLORS[GridConstants.Wall],
#         )
#         wall_img = np.moveaxis(wall_img, -1, 0)  # need channel dim to be last for the above fill to work
#         np.testing.assert_array_equal(
#             self.env.world_map_color[:, pad, pad:-pad],
#             wall_img[:, pad, pad:-pad],
#         )
#         np.testing.assert_array_equal(
#             self.env.world_map_color[:, -pad - 1, pad:-pad],
#             wall_img[:, -pad - 1, pad:-pad],
#         )
#         np.testing.assert_array_equal(
#             self.env.world_map_color[:, pad:-pad, pad],
#             wall_img[:, pad:-pad, pad],
#         )
#         np.testing.assert_array_equal(
#             self.env.world_map_color[:, pad:-pad, -pad - 1],
#             wall_img[:, pad:-pad, -pad - 1],
#         )
#
#     def assert_logical_and_image_view(self, agent_id, expected_logical_view):
#         obs = self.env.get_obs()
#         image_view = obs[agent_id]["image_view"]
#         logical_view = obs[agent_id]["ascii_view"]
#
#         expected_image_view = map_to_image(expected_logical_view)
#
#         np.testing.assert_array_equal(logical_view, expected_logical_view)
#         np.testing.assert_array_equal(image_view, expected_image_view)
#
#     def test_view(self):
#         self._construct_map(BASE_MAP, num_agents=1, start_positions=[(3, 3)], view_size=2)
#         a_id = self.env.agent_ids[0]
#
#         # Test that the view is empty around the agent
#         expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "     ", "     "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Move agent one row closer to the top, check that the top row is walls
#         self.change_agent_position(a_id, (2, 3))
#         expected_logical_view = ascii_to_numpy(["#####", "     ", "  1  ", "     ", "     "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check that the agent sees that padded view when the row above them is out of view
#         self.change_agent_position(a_id, (1, 3))
#         expected_logical_view = ascii_to_numpy(["     ", "#####", "  1  ", "     ", "     "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check that the view is correct if the left wall is in view
#         self.change_agent_position(a_id, (3, 2))
#         expected_logical_view = ascii_to_numpy(["#    ", "#    ", "# 1  ", "#    ", "#    "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if it exceeds the left view
#         self.change_agent_position(a_id, (3, 1))
#         expected_logical_view = ascii_to_numpy([" #   ", " #   ", " #1  ", " #   ", " #   "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if the bottom is in view
#         self.change_agent_position(a_id, (4, 3))
#         expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "     ", "#####"])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if exceeding bottom view
#         self.change_agent_position(a_id, (5, 3))
#         expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "#####", "     "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if the right wall is in view
#         self.change_agent_position(a_id, (3, 4))
#         expected_logical_view = ascii_to_numpy(["    #", "    #", "  1 #", "    #", "    #"])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if exceeding the right wall
#         self.change_agent_position(a_id, (3, 5))
#         expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "   # ", "   # "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#         # Check if the view is correct if in the bottom right corner
#         self.change_agent_position(a_id, (5, 5))
#         expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "#### ", "     "])
#         self.assert_logical_and_image_view(a_id, expected_logical_view)
#
#     def test_agent_visibility(self):
#         self._construct_map(BASE_MAP.copy(), 0, None, view_size=2, obs=["image_view", "other_agent_visibility"])
#         self.add_agent_to_env("agent-0", (1, 1))
#         self.add_agent_to_env("agent-1", (1, 3))
#         self.add_agent_to_env("agent-2", (1, 5))
#         obs, *_ = self.env.step({})
#         visibility = [a_obs["other_agent_visibility"] for a_obs in obs.values()]
#         np.testing.assert_array_equal(visibility[0], [1, 0])
#         np.testing.assert_array_equal(visibility[1], [1, 1])
#         np.testing.assert_array_equal(visibility[2], [0, 1])
#
#     def test_agent_actions(self):
#         a_id = "agent-0"
#         pad = 2
#         self._construct_map(
#             BASE_MAP.copy(),
#             1,
#             [(2, 2)],
#             pad,
#         )
#
#         self.env.step({a_id: BASE_ACTION_MAP["NOOP"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
#         np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")
#         np.testing.assert_array_equal(self.env.ascii_map[pad + 2, pad + 2], "1")
#         np.testing.assert_array_equal(self.env.world_map_color[:, pad + 2, pad + 2], DEFAULT_COLORS["1"])
#
#         self.env.step({a_id: BASE_ACTION_MAP["LEFT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 1])
#         np.testing.assert_array_equal(self.env.map_with_agents[2, 1], "1")
#         np.testing.assert_array_equal(self.env.ascii_map[pad + 2, pad + 1], "1")
#         np.testing.assert_array_equal(self.env.world_map_color[:, pad + 2, pad + 1], DEFAULT_COLORS["1"])
#
#         self.env.step({a_id: BASE_ACTION_MAP["RIGHT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
#         np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")
#         np.testing.assert_array_equal(self.env.ascii_map[pad + 2, pad + 2], "1")
#         np.testing.assert_array_equal(self.env.world_map_color[:, pad + 2, pad + 2], DEFAULT_COLORS["1"])
#
#         self.env.step({a_id: BASE_ACTION_MAP["UP"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [1, 2])
#         np.testing.assert_array_equal(self.env.map_with_agents[1, 2], "1")
#         np.testing.assert_array_equal(self.env.ascii_map[pad + 1, pad + 2], "1")
#         np.testing.assert_array_equal(self.env.world_map_color[:, pad + 1, pad + 2], DEFAULT_COLORS["1"])
#
#         self.env.step({a_id: BASE_ACTION_MAP["DOWN"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
#         np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")
#         np.testing.assert_array_equal(self.env.ascii_map[pad + 2, pad + 2], "1")
#         np.testing.assert_array_equal(self.env.world_map_color[:, pad + 2, pad + 2], DEFAULT_COLORS["1"])
#
#         # if an agent tries to move through a wall they should stay in the same place
#         # we check that this works correctly for both corner and non-corner edges
#         self.change_agent_position(a_id, (1, 1))
#         self.env.step({a_id: BASE_ACTION_MAP["UP"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 1))
#
#         self.env.step({a_id: BASE_ACTION_MAP["LEFT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 1))
#
#         self.change_agent_position(a_id, (4, 4))
#         self.env.step({a_id: BASE_ACTION_MAP["RIGHT"]})
#         self.env.step({a_id: BASE_ACTION_MAP["DOWN"]})
#         self.env.step({a_id: BASE_ACTION_MAP["RIGHT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 5))
#
#         self.env.step({a_id: BASE_ACTION_MAP["DOWN"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 5))
#
#         self.env.step({a_id: BASE_ACTION_MAP["LEFT"]})
#         self.env.step({a_id: BASE_ACTION_MAP["DOWN"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 4))
#
#         self.change_agent_position(a_id, (4, 5))
#         self.env.step({a_id: BASE_ACTION_MAP["RIGHT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (4, 5))
#
#         self.change_agent_position(a_id, (2, 1))
#         self.env.step({a_id: BASE_ACTION_MAP["LEFT"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (2, 1))
#
#         self.change_agent_position(a_id, (1, 2))
#         self.env.step({a_id: BASE_ACTION_MAP["UP"]})
#         np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 2))
#
#     def test_agent_conflict(self):
#         self._construct_map(BASE_MAP_2.copy(), 2, [(1, 2), (4, 4)])
#
#         # test that agents can't walk into other agents
#         self.change_agent_position("agent-0", (3, 3))
#         self.change_agent_position("agent-1", (3, 4))
#         self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"]})
#         self.env.step({"agent-1": BASE_ACTION_MAP["LEFT"]})
#         np.testing.assert_array_equal(self.env.env_agents["agent-0"].pos, (3, 3))
#         np.testing.assert_array_equal(self.env.env_agents["agent-1"].pos, (3, 4))
#
#         # test that agents can't walk through each other if they move simultaneously
#         self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["LEFT"]})
#         np.testing.assert_array_equal(self.env.env_agents["agent-0"].pos, (3, 3))
#         np.testing.assert_array_equal(self.env.env_agents["agent-1"].pos, (3, 4))
#         # also check that the map looks correct, no agent has disappeared
#         expected_map = ascii_to_numpy(["######", "#    #", "#    #", "#  12#", "#    #", "######"])
#
#         np.testing.assert_array_equal(expected_map, self.env.map_with_agents)
#
#         # test that agents can walk into other agents if moves are de-conflicting
#         # conflict only occurs stochastically so try it 50 times
#         np.random.seed(1)
#         self.change_agent_position("agent-0", (2, 4))
#         self.change_agent_position("agent-1", (3, 4))
#         expected_map = ascii_to_numpy(["######", "#    #", "#   1#", "#   2#", "#    #", "######"])
#         self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["UP"]})
#         np.testing.assert_array_equal(expected_map, self.env.map_with_agents)
#         self.env.step({"agent-0": BASE_ACTION_MAP["LEFT"], "agent-1": BASE_ACTION_MAP["DOWN"]})
#         self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["UP"]})
#         np.testing.assert_array_equal(expected_map, self.env.map_with_agents)
#
#
# if __name__ == "__main__":
#     unittest.main()



from __future__ import annotations

import dataclasses


@dataclasses.dataclass
class Actions:
    # Common
    PickupDrop = "pick_up_or_drop"  # Pick up an object
    Toggle = "toggle"  # Interact with/toggle/activate an object
    Noop = "no-op"

    # Rotation
    Forward = "move_forward"  # Move forward
    RotateLeft = "rotate_left"  # Turn Left
    RotateRight = "rotate_right"  # Turn Right

    # Cardinal Movement
    MoveLeft = "move_left"
    MoveRight = "move_right"
    MoveUp = "move_up"
    MoveDown = "move_down"


@dataclasses.dataclass
class ActionSets:
    RotationActions = (
        Actions.Forward,
        Actions.PickupDrop,
        Actions.Toggle,
        Actions.Noop,
        Actions.RotateLeft,
        Actions.RotateRight,
    )
    CardinalActions = (
        Actions.MoveUp,
        Actions.MoveDown,
        Actions.MoveLeft,
        Actions.MoveRight,
        Actions.PickupDrop,
        Actions.Toggle,
        Actions.Noop,
    )



import copy

import numpy as np

from cogrid.core import directions
from cogrid import cogrid_env
from cogrid.envs.overcooked import agent
from cogrid.envs import registry
from cogrid.envs.overcooked import rewards


class Overcooked(cogrid_env.CoGridEnv):
    """The Overcooked task is a reproduction of the Overcooked-AI
    environment from Carroll et al. (2019). The original paper
    can be found here: https://arxiv.org/abs/1910.05789 and
    source code is provided through
    https://github.com/HumanCompatibleAI/overcooked_ai
    """

    def __init__(self, config, render_mode=None, **kwargs):
        """Constructor method"""
        super().__init__(
            config=config,
            render_mode=render_mode,
            agent_class=agent.OvercookedAgent,
            **kwargs,
        )



"""GridObj defines an object in the CoGridEnv environment. It is largely derived from the Minigrid WorldObj:
https://github.com/Farama-Foundation/Minigrid/minigrid/core/world_object.py

"""

from __future__ import annotations
from copy import deepcopy
import math
import uuid

import numpy as np


from cogrid.constants import GridConstants
from cogrid.core.constants import ObjectColors, Colors, COLORS, COLOR_NAMES
from cogrid.core import constants
from cogrid.core.directions import Directions
from cogrid.visualization.rendering import (
    fill_coords,
    point_in_circle,
    point_in_rect,
    point_in_triangle,
    rotate_fn,
    add_text_to_image,
)


# The OBJECT_REGISTRY holds all registered objects under a "scope" (e.g., "global", "search_rescue", "overcooked")
# which allows us to re-use IDs and character representations across environments (e.g., P is a purple target in
# search_rescue and a plate in overcooked).
OBJECT_REGISTRY: dict[str, dict[str, GridObj]] = {}


def make_object(
    object_id: str | None, scope: str = "global", **kwargs
) -> GridObj:
    if object_id is None:
        return None

    if scope not in OBJECT_REGISTRY:
        raise ValueError(
            f"No objects registered with scope `{scope}`. Existing scopes are {list(OBJECT_REGISTRY.keys())}."
        )

    if object_id in OBJECT_REGISTRY["global"]:
        return OBJECT_REGISTRY["global"][object_id](**kwargs)
    elif object_id not in OBJECT_REGISTRY[scope]:
        raise ValueError(
            f"Object with object_id `{object_id}` not registered in scope `{scope}`. "
            f"Call register_object('{object_id}', <class>, scope='{scope}') to add it to the registry."
        )

    return OBJECT_REGISTRY[scope][object_id](**kwargs)


def get_object_class(object_id: str, scope: str = "global") -> GridObj:
    return OBJECT_REGISTRY[scope][object_id]


def register_object(
    object_id: str, obj_class: GridObj, scope: str = "global"
) -> None:
    global_scope_chars = [
        obj.char for obj in OBJECT_REGISTRY.get("global", {}).values()
    ]

    if obj_class.char in global_scope_chars:
        raise ValueError(
            f"Character `{obj_class.char}` is already in use in the global scope. "
            "Please choose a different character."
        )

    if object_id in OBJECT_REGISTRY.get("global", {}):
        raise ValueError(
            f"Object with object_id `{object_id}` already registered in the global scope. "
            "Please select a different ID."
        )

    if scope not in OBJECT_REGISTRY:
        OBJECT_REGISTRY[scope] = {}

    OBJECT_REGISTRY[scope][object_id] = obj_class


def get_registered_object_ids(scope: str = "global") -> list[str]:
    """Return a list of the object_ids of available objects in a given scope."""
    return list(OBJECT_REGISTRY[scope].keys())


def get_object_char(object_id: str, scope: str = "global") -> str:
    return get_object_class(object_id, scope=scope).char


def get_object_id_from_char(object_char: str, scope: str = "global") -> str:
    # First check global scope, no matter what scope was passed (default to global).
    for object_id, object_class in OBJECT_REGISTRY["global"].items():
        if object_class.char == object_char:
            return object_id

    if scope != "global":
        for object_id, object_class in OBJECT_REGISTRY[scope].items():
            if object_class.char == object_char:
                return object_id

    raise ValueError(
        f"There is no registered object with character representation `{object_char}` in scope `{scope}`."
    )


# TODO(chase): This is temporary and needs to be retrieved from the registry.
OBJECT_NAMES = [
    None,
    "free_space",
    "wall",
    "door",
    "key",
    "pickaxe",
    "medkit",
    "rubble",
    "green_victim",
    "purple_victim",
    "yellow_victim",
    "red_victim",
    "pot",
    "plate",
    "plate_stack",
    "onion_soup",
    "onion_stack",
    "onion",
    "delivery_zone",
    "counter",
] + [f"agent_{direction}" for direction in "^>v<"]


class GridObj:
    object_id: str = None
    color: str | tuple = None
    char: str = None

    def __init__(
        self,
        state: int = 0,
        toggle_value: float = 0,
        inventory_value: float = 0,
        overlap_value: float = 0,
        placed_on_value: float = 0,
        picked_up_from_value: float = 0,
    ):
        self.uuid: str = str(uuid.uuid4())

        self.state: int = state

        # If an object can be placed on top of this one, this will hold the object that's on top.
        self.obj_placed_on: GridObj | None = None

        # position info
        self.init_pos: tuple[int, int] | None = None
        self.pos: tuple[int, int] | None = None

        # defines rewards for holding/toggling/overlapping
        self.toggle_value: float | int = toggle_value
        self.inventory_value: float | int = inventory_value
        self.overlap_value: float | int = overlap_value
        self.placed_on_value: float | int = placed_on_value
        self.picked_up_from_value: float | int = picked_up_from_value

    def can_overlap(self, agent: GridAgent) -> bool:
        """Can an agent overlap with this object?"""
        return False

    def can_pickup(self, agent: GridAgent) -> bool:
        """Can an agent pick this object up and store in inventory?"""
        return False

    def can_place_on(self, agent: GridAgent, cell: GridObj) -> bool:
        """
        Can another object be placed on top of this object? e.g., a countertop that can't be walked through
        but can have an item on top of it.
        """
        return False

    def can_pickup_from(self, agent: GridAgent) -> bool:
        """Can the agent pick up an object from this one?"""
        return self.obj_placed_on is not None and self.obj_placed_on.can_pickup(
            agent=agent
        )

    def place_on(self, agent: GridAgent, cell: GridObj) -> None:
        self.obj_placed_on = cell
        self.state = object_to_idx(cell)

    def pick_up_from(self, agent: GridAgent) -> GridObj:
        assert (
            self.obj_placed_on is not None
        ), f"Picking up from but there's no object placed on {self.object_id}"
        cell = self.obj_placed_on
        self.obj_placed_on = None
        self.state = 0
        return cell

    def see_behind(self, agent: GridAgent) -> bool:
        """Can the agent see through this object?"""
        return True

    def visible(self) -> bool:
        return True

    def toggle(self, env, agent: GridAgent = None) -> bool:
        """
        Trigger/Toggle an action this object performs. Some toggles are conditioned on the environment
        and require specific conditions to be met, which can be checked with the end.
        """
        return False

    def encode(self, encode_char=True):
        return (
            self.char if encode_char else object_to_idx(self),
            0,  # TODO(chase): Remove. This used to be color, but we're no longer using it.
            int(self.state),
        )

    def render(self, tile_img):
        """By default, everything will be rendered as a square with the specified color."""
        fill_coords(tile_img, point_in_rect(0, 1, 0, 1), color=self.color)

    @staticmethod
    def decode(char_or_idx: str | int, state: int, scope: str = "global"):

        if char_or_idx in [
            None,
            GridConstants.FreeSpace,
            GridConstants.Obscured,
        ]:
            return None

        # check if the name was passed instead of the character
        if _is_str(char_or_idx) and len(char_or_idx) > 1:
            object_id = char_or_idx
        elif _is_str(char_or_idx):
            object_id = get_object_id_from_char(char_or_idx, scope=scope)
        else:
            raise ValueError(f"Invalid identifier for decoding: {char_or_idx}")

        state = int(state)

        return make_object(object_id, state=state, scope=scope)

    def rotate_left(self):
        """Some objects (e.g., agents) have a rotation and must be rotated with the grid."""
        pass

    def tick(self):
        """
        Some objects have a time component (e.g., cooking soup), so we call the tick
        method on all objects for each env.step()
        """
        pass

    def _remove_from_grid(self, grid):
        cell = grid.get(*self.pos)
        assert self is cell
        grid.set(*self.pos, None)


def _is_str(chk):
    return isinstance(chk, str) or isinstance(chk, np.str)


def _is_int(chk):
    return isinstance(chk, int) or isinstance(chk, np.int)


def object_to_idx(object: GridObj | str | None):
    if isinstance(object, GridObj):
        object_id = object.object_id
    else:
        object_id = object

    return OBJECT_NAMES.index(object_id)


def idx_to_object(idx: int):
    for name in OBJECT_NAMES:
        if OBJECT_NAMES.index(name) == idx:
            return name
    raise ValueError(f"Object index {idx} not in OBJECT_LIST.")


class GridAgent(GridObj):
    def __init__(self, agent):
        """
        Grid agents are initialized slightly differently. State corresponds to the object they are holding
        and char/colors are unique for each agent.
        """
        self.color = {
            1: ObjectColors.AgentOne,
            2: ObjectColors.AgentTwo,
            3: ObjectColors.AgentThree,
            4: ObjectColors.AgentFour,
        }[agent.agent_number]

        self.char = {
            Directions.Up: "^",
            Directions.Down: "v",
            Directions.Left: "<",
            Directions.Right: ">",
        }[agent.dir]

        assert (
            len(agent.inventory) <= 1
        ), "Current implementation requires maximum inventory size of 1."

        self.object_id = f"agent_{self.char}"

        # TODO(chase): State must encapsulate carried objects and role
        # state = agent.role_idx
        state = (
            0
            if len(agent.inventory) == 0
            else object_to_idx(agent.inventory[0])
        )

        super().__init__(state=state)
        self.dir = agent.dir
        self.pos = agent.pos
        self.front_pos = agent.front_pos
        self.agent_id = agent.id
        self.inventory: list[GridObj] = deepcopy(agent.inventory)
        assert self.pos is not None

    def rotate_left(self):
        self.char = {"^": "<", "<": "v", "v": ">", ">": "^"}[self.char]
        self.object_id = f"agent_{self.char}"
        self.dir -= 1
        if self.dir < 0:
            self.dir += 4

    def render(self, tile_img):
        tri_fn = point_in_triangle(
            (0.12, 0.19),
            (0.87, 0.50),
            (0.12, 0.81),
        )

        # Rotate the triangle based on agent direction
        assert self.dir is not None
        tri_fn = rotate_fn(
            tri_fn, cx=0.5, cy=0.5, theta=0.5 * math.pi * self.dir
        )
        fill_coords(tile_img, tri_fn, self.color)

        # add any item in the inventory to the corner
        inv_tile_rows, inv_tile_cols = (
            tile_img.shape[0] // 3,
            tile_img.shape[1] // 3,
        )
        assert (
            len(self.inventory) <= 3
        ), "We're rendering inventory items at 1/3 size, so can't do more than 3!"

        offset = 4  # offset so we still see grid lines
        for i, obj in enumerate(self.inventory):
            inventory_tile = np.zeros(shape=(inv_tile_rows, inv_tile_cols, 3))
            obj.render(inventory_tile)

            # Take the subset of the image that we'll fill, then only fill where the image is non-zero
            # this makes a transparent background, rather than adding in a black square
            tile_subset = tile_img[
                i * inv_tile_rows + offset : (i + 1) * inv_tile_rows + offset,
                offset : inv_tile_cols + offset,
                :,
            ]
            nonzero_entries = np.nonzero(inventory_tile)
            tile_subset[nonzero_entries] = inventory_tile[nonzero_entries]

    @staticmethod
    def decode(char_or_idx: str | int, state: int):

        if char_or_idx in [
            None,
            GridConstants.FreeSpace,
            GridConstants.Obscured,
        ]:
            return None

        # check if the name was passed instead of the character
        if _is_str(char_or_idx) and len(char_or_idx) > 1:
            object_id = char_or_idx
        elif _is_str(char_or_idx):
            object_id = get_object_id_from_char(char_or_idx)
        elif _is_int(char_or_idx):
            object_id = OBJECT_NAMES[char_or_idx]
        else:
            raise ValueError(f"Invalid identifier for decoding: {char_or_idx}")

        state = int(state)

        return make_object(object_id, state=state)


class Wall(GridObj):
    object_id = "wall"
    color = constants.Colors.Grey
    char = "#"

    def __init__(self, *args, **kwargs):
        super().__init__(state=0)

    def see_behind(self) -> bool:
        return False


register_object(Wall.object_id, Wall, scope="global")


class Floor(GridObj):
    object_id = "floor"
    color = constants.Colors.PaleBlue
    char = GridConstants.FreeSpace

    def __init__(self, **kwargs):
        super().__init__(
            state=0,
        )

    def can_overlap(self) -> bool:
        return True


register_object(Floor.object_id, Floor, scope="global")


class Counter(GridObj):
    object_id = "counter"
    color = constants.Colors.LightBrown
    char = "C"

    def __init__(self, state: int = 0, **kwargs):
        # TODO(chase): need to be able to initialize an object on top
        #   via the state. Take state and map it to an object.
        super().__init__(
            state=state,
        )

    def can_place_on(self, agent: GridAgent, cell: GridObj) -> bool:
        return self.state == 0

    def render(self, tile_img):
        super().render(tile_img)

        if self.obj_placed_on is not None:
            self.obj_placed_on.render(tile_img)


register_object(Counter.object_id, Counter, scope="global")


class Key(GridObj):
    object_id = "key"
    color = constants.Colors.Yellow
    char = "K"

    def __init__(self, state=0):
        super().__init__(state=state)

    def can_pickup(self, agent: GridAgent):
        return True

    def render(self, tile_img):
        # Vertical quad
        fill_coords(tile_img, point_in_rect(0.50, 0.63, 0.31, 0.88), self.color)

        # Teeth
        fill_coords(tile_img, point_in_rect(0.38, 0.50, 0.59, 0.66), self.color)
        fill_coords(tile_img, point_in_rect(0.38, 0.50, 0.81, 0.88), self.color)

        # Ring
        fill_coords(
            tile_img, point_in_circle(cx=0.56, cy=0.28, r=0.190), self.color
        )
        fill_coords(
            tile_img, point_in_circle(cx=0.56, cy=0.28, r=0.064), (0, 0, 0)
        )


register_object(Key.object_id, Key, scope="global")


class Door(GridObj):
    object_id = "door"
    color = constants.Colors.DarkGrey
    char = "D"

    def __init__(self, state):
        super().__init__(state=state)
        self.is_open = state == 2
        self.is_locked = state == 0

    def can_overlap(self, agent: GridAgent) -> bool:
        """The agent can only walk over this cell when the door is open"""
        return self.is_open

    def see_behind(self, agent: GridAgent) -> bool:
        return self.is_open

    def toggle(self, env, agent: GridAgent) -> bool:
        if self.is_locked:
            if any([isinstance(obj, Key) for obj in agent.inventory]):
                self.is_locked = False
                self.is_open = True
                return True
            return False

        self.is_open = not self.is_open
        return True

    def encode(self, encode_char=False):
        """Encode the a description of this object as a 3-tuple of integers"""

        # State, 0: open, 1: closed, 2: locked
        if self.is_open:
            self.state = 2
        elif self.is_locked:
            self.state = 0
        # if door is closed and unlocked
        elif not self.is_open:
            self.state = 1
        else:
            raise ValueError(
                f"There is no possible state encoding for the state:\n -Door Open: {self.is_open}\n -Door Closed: {not self.is_open}\n -Door Locked: {self.is_locked}"
            )

        return super().encode(encode_char=encode_char)

    def render(self, tile_img):

        if self.state == 2:
            fill_coords(
                tile_img, point_in_rect(0.88, 1.00, 0.00, 1.00), self.color
            )
            fill_coords(
                tile_img, point_in_rect(0.92, 0.96, 0.04, 0.96), (0, 0, 0)
            )
            return

        # Door frame and door
        if self.state == 0:
            fill_coords(
                tile_img, point_in_rect(0.00, 1.00, 0.00, 1.00), self.color
            )
            fill_coords(
                tile_img,
                point_in_rect(0.06, 0.94, 0.06, 0.94),
                0.45 * np.array(self.color),
            )

            # Draw key slot
            fill_coords(
                tile_img, point_in_rect(0.52, 0.75, 0.50, 0.56), self.color
            )
        else:
            fill_coords(
                tile_img, point_in_rect(0.00, 1.00, 0.00, 1.00), self.color
            )
            fill_coords(
                tile_img, point_in_rect(0.04, 0.96, 0.04, 0.96), (0, 0, 0)
            )
            fill_coords(
                tile_img, point_in_rect(0.08, 0.92, 0.08, 0.92), self.color
            )
            fill_coords(
                tile_img, point_in_rect(0.12, 0.88, 0.12, 0.88), (0, 0, 0)
            )

            # Draw door handle
            fill_coords(
                tile_img, point_in_circle(cx=0.75, cy=0.50, r=0.08), self.color
            )


register_object(Door.object_id, Door, scope="global")



import numpy as np

from cogrid.core.directions import Directions
from cogrid.core.grid_object import GridObj


class Agent:
    def __init__(self, agent_id, start_position, start_direction, **kwargs):
        """
        When developing a subclass, you must, at the very least, implement `consume()`. This
        defines how an agent interacts with their current position on the map and sets their reward
        for that turn.

        :param agent_id: unique identifier
        :param start_position: starting spawn position
        :param env: 2d grid environment
        :param obs_type: `position' or 'view'. Position simply returns (row, col) coordinates,
                view returns a matrix representing the FoV.
        :param image: (boolean) If true, a `view' observation will return an RGB image (e.g., for CNNs)
        :param row_view: how many rows in either direction the agent can view
        :param col_view: how many columns in either direction the agent can view
        :param actions: dict keyed by integers with corresponding string actions
        """
        self.id: str = agent_id
        self.pos: tuple[int, int] = start_position
        self.dir: Directions = start_direction
        self.role: str = None
        self.role_idx: int = None
        self.inventory_capacity: int = kwargs.get("inventory_capacity", 1)

        self.terminated: bool = False

        self.collision: bool = (
            False  # Some envs keep track of if an agent crashed into another agent/object/etc.
        )

        self.orientation: str = "down"
        self.inventory: list[GridObj] = []
        self.cell_toggled: GridObj | None = None
        self.cell_placed_on: GridObj | None = None
        self.cell_picked_up_from: GridObj | None = None
        self.cell_overlapped: GridObj | None = None

    def rotate_left(self):
        self.dir -= 1
        if self.dir < 0:
            self.dir += 4

    def rotate_right(self):
        self.dir = (self.dir + 1) % 4

    @property
    def front_pos(self):
        return self.pos + self.dir_vec

    @property
    def dir_vec(self):
        # dir_to_vec = [
        #     # Pointing right (positive X)
        #     np.array((0, 1)),
        #     # Down (positive Y)
        #     np.array((1, 0)),
        #     # Pointing left (negative X)
        #     np.array((0, -1)),
        #     # Up (negative Y)
        #     np.array((-1, 0)),
        # ]
        dir_to_vec = {
            Directions.Right: np.array((0, 1)),  # Increase col away from 0
            Directions.Down: np.array(
                (1, 0)
            ),  # Down increases the row number (0 is top)
            Directions.Left: np.array(
                (0, -1)
            ),  # Left decreases the col towards 0
            Directions.Up: np.array(
                (-1, 0)
            ),  # Up decreases the row to 0 (move towards the top)
        }
        return dir_to_vec[self.dir]

    @property
    def right_vec(self):
        # TODO(chase): check that this is correct
        dy, dx = self.dir_vec
        return np.array((dx, -dy))

    def set_orientation(self):
        self.orientation = {
            Directions.Up: "up",
            Directions.Down: "down",
            Directions.Left: "left",
            Directions.Right: "right",
        }[self.dir]

    def can_pickup(self, grid_object: GridObj) -> bool:
        return len(self.inventory) < self.inventory_capacity

    def pick_up_object(self, grid_object: GridObj):
        self.inventory.append(grid_object)

    @property
    def agent_number(self) -> int:
        """Converts agent id to integer, beginning with 1,
        e.g., agent-0 -> 1, agent-1 -> 2, etc.
        """
        return int(self.id[-1]) + 1 if isinstance(self.id, str) else self.id + 1



"""Feature generators takes the CoGridEnv object and turns it into the desired observation"""

import numpy as np
from gymnasium import spaces
from collections import deque

from cogrid.core.grid_object import OBJECT_NAMES
from cogrid.feature_space import feature, feature_space
from cogrid.core import grid_utils

try:
    import cv2
except ImportError:
    cv2 = None


class FullMapImage(feature.Feature):
    def __init__(self, map_size, tile_size, **kwargs):
        rows, cols = map_size
        super().__init__(
            low=0,
            high=1,
            shape=(rows * tile_size, cols * tile_size, 3),
            name="full_map_image",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        img = env.get_full_render(highlight=False)
        img = (img / 255.0).astype(np.float32)
        return img


feature_space.register_feature("full_map_image", FullMapImage)


class StackedFullMapResizedGrayscale(feature.Feature):
    def __init__(self, **kwargs):
        assert (
            cv2 is not None
        ), "Must install cv2 to use image resizing. Run `pip install opencv-python` then try again."
        super().__init__(
            low=0,
            high=1,
            shape=(84, 84, 4),
            name="stacked_full_map_resized_grayscale_image",
            **kwargs
        )
        self.frames = deque(maxlen=4)
        for _ in range(4):
            self.frames.append(np.zeros((84, 84, 1)))

        self.player_id = None

    def generate(self, env, player_id, **kwargs):
        if self.player_id is not None:
            assert player_id == self.player_id
        else:
            self.player_id = player_id

        img_rgb = env.get_full_render(highlight=False)

        assert img_rgb.shape[-1] == 3

        img_resized = cv2.resize(
            img_rgb, (84, 84), interpolation=cv2.INTER_AREA
        )
        img_grayscale = cv2.cvtColor(img_resized, cv2.COLOR_RGB2GRAY)
        img_grayscale = np.expand_dims(img_grayscale, -1)
        self.frames.append(img_grayscale / 255.0)

        stacked = np.stack(self.frames, axis=-1).reshape(self.shape)

        return stacked


feature_space.register_feature(
    "stacked_full_map_resized_grayscale_image",
    StackedFullMapResizedGrayscale,
)


class FullMapResizedGrayscale(feature.Feature):
    def __init__(self, shape=(42, 42, 1), **kwargs):
        super().__init__(
            low=0,
            high=1,
            shape=shape,
            name="full_map_resized_grayscale_image",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        img_rgb = env.get_full_render(highlight=False)
        img_resized = cv2.resize(
            img_rgb, self.shape[0:2], interpolation=cv2.INTER_AREA
        )
        img_grayscale = cv2.cvtColor(img_resized, cv2.COLOR_RGB2GRAY)
        img_grayscale = np.expand_dims(img_grayscale, -1)

        return img_grayscale / 255.0


feature_space.register_feature(
    "full_map_resized_grayscale_image",
    FullMapResizedGrayscale,
)


class FoVImage(feature.Feature):
    def __init__(self, view_len, tile_size, **kwargs):
        super().__init__(
            low=0,
            high=1,
            shape=(view_len * tile_size, view_len * tile_size, 3),
            name="fov_image",
            **kwargs
        )
        self.view_len = view_len

    def generate(self, env, player_id, **kwargs):
        img = env.get_pov_render(agent_id=player_id)
        img = (img / 255.0).astype(np.float32)
        return img


feature_space.register_feature("fov_image", FoVImage)


class FullMapEncoding(feature.Feature):
    max_map_size = (12, 12)

    def __init__(self, **kwargs):
        # TODO(chase): We need to determine a high value for the encodings
        super().__init__(
            low=0,
            high=np.inf,
            shape=(*self.max_map_size, 3),
            name="full_map_encoding",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.uint8)
        encoded_map = env.grid.encode(encode_char=False)

        # Fill out the encoding with the encoded map, the encoding will be 0 padded if the map is smaller than the max map size
        encoding[: encoded_map.shape[0], : encoded_map.shape[1], :] = (
            encoded_map
        )
        return encoding


feature_space.register_feature("full_map_encoding", FullMapEncoding)


class FoVEncoding(feature.Feature):
    def __init__(self, view_len, **kwargs):
        # TODO(chase): We need to determine a high value for the encodings
        super().__init__(
            low=0,
            high=100,
            shape=(view_len, view_len, 3),
            name="fov_encoding",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        agent_grid, _ = env.gen_obs_grid(agent_id=player_id)
        encoded_agent_grid = agent_grid.encode(encode_char=False)
        return encoded_agent_grid


feature_space.register_feature("fov_encoding", FoVEncoding)


class FullMapASCII(feature.Feature):
    def __init__(self, map_size, **kwargs):
        super().__init__(
            low=-np.inf,
            high=np.inf,
            shape=(*map_size, 3),
            name="full_map_ascii",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        encoded_map = env.grid.encode(encode_char=True)
        return encoded_map


feature_space.register_feature("full_map_ascii", FullMapASCII)


class FoVASCII(feature.Feature):
    def __init__(self, view_len, **kwargs):
        super().__init__(
            low=-np.inf,
            high=np.inf,
            shape=(view_len, view_len, 3),
            name="fov_ascii",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        agent_grid, _ = env.gen_obs_grid(agent_id=player_id)
        encoded_agent_grid = agent_grid.encode(encode_char=True)

        # TODO(chase): Confirm that this shouldn't already be correct
        # assert encoded_agent_grid[0, -1, agent_grid.width // 2] == "^"
        encoded_agent_grid[0, -1, agent_grid.width // 2] = "^"

        return encoded_agent_grid


feature_space.register_feature("fov_ascii", FoVASCII)


class AgentPosition(feature.Feature):

    def __init__(self, **kwargs):
        super().__init__(
            low=0, high=np.inf, shape=(2,), name="agent_position", **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        return np.asarray(env.env_agents[player_id].pos, dtype=np.int32)


feature_space.register_feature("agent_position", AgentPosition)


class AgentPositions(feature.Feature):
    def __init__(self, map_shape, **kwargs):
        self.rgb = True
        super().__init__(
            low=0,
            high=1,
            shape=(
                *map_shape,
                1 if not self.rgb else 3,
            ),
            name="agent_positions",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        channel_dim = 1 if not self.rgb else 3
        grid = np.full((*env.map_with_agents.shape, channel_dim), fill_value=0)
        for a_id, agent in env.env_agents.items():
            if (
                agent is not None
            ):  # will be None before being set by subclassed env
                assert not self.rgb, "RGB not implemented for new grid."
                # if self.rgb:
                #     grid[:, agent.pos[0], agent.pos[1]] = (
                #         np.array(constants.DEFAULT_COLORS[str(env.id_to_numeric(a_id))]) / 255.0
                #     )
                # else:
                grid[:, agent.pos[0], agent.pos[1]] = int(
                    env.id_to_numeric(a_id)
                )

        return grid


feature_space.register_feature("agent_positions", AgentPositions)


class AgentDir(feature.Feature):
    """One-hot encoding of the agent's direction."""

    def __init__(self, **kwargs):
        super().__init__(low=0, high=1, shape=(4,), name="agent_dir")

    def generate(self, env, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.int32)
        encoding[env.env_agents[player_id].dir] = 1
        return encoding


feature_space.register_feature("agent_dir", AgentDir)


class OtherAgentActions(feature.Feature):
    def __init__(self, num_agents, num_actions, **kwargs):
        super().__init__(
            low=0,
            high=num_actions,
            shape=(num_actions * (num_agents - 1),),
            name="other_agent_actions",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        return (
            np.array(
                [
                    self.one_hot_encode_actions(
                        env.prev_actions[a_id], self.high
                    )
                    for a_id in env.agent_ids
                    if a_id is not player_id
                ]
            )
            .reshape(-1)
            .astype(np.uint8)
        )

    @staticmethod
    def one_hot_encode_actions(action, num_actions):
        oh_action = np.zeros(num_actions)
        oh_action[action] = 1
        return oh_action


feature_space.register_feature("other_agent_actions", OtherAgentActions)


class OtherAgentVisibility(feature.Feature):
    def __init__(self, num_agents, view_len, **kwargs):
        super().__init__(
            low=0,
            high=1,
            shape=(num_agents - 1,),
            name="other_agent_visibility",
            **kwargs
        )
        self.view_len = view_len
        self.num_other_agents = num_agents - 1

    def generate(self, env, player_id, **kwargs):
        raise NotImplementedError
        # agent = env.env_agents[player_id]
        # view = ascii_view(env.ascii_map, agent.pos, self.view_len)
        # visibility = np.zeros((len(env.agent_ids) - 1,))
        # other_agent_ids = [pid for pid in env.agent_ids if pid != player_id]
        # for i, other_agent_id in enumerate(other_agent_ids):
        #     numeric_id = env.id_to_numeric(other_agent_id)
        #     visibility[i] = int(numeric_id in view)
        # return visibility


feature_space.register_feature("other_agent_visibility", OtherAgentVisibility)


class Role(feature.Feature):
    def __init__(self, num_roles, **kwargs):
        super().__init__(
            low=0, high=num_roles - 1, shape=(num_roles,), name="role", **kwargs
        )
        self.num_roles = num_roles

    def generate(self, env, player_id, **kwargs):
        agent = env.env_agents[player_id]
        role_encoding = np.zeros((self.num_roles,), dtype=np.uint8)
        role_encoding[agent.role_idx] = 1
        return role_encoding


feature_space.register_feature("role", Role)


class Inventory(feature.Feature):
    def __init__(self, inventory_capacity, **kwargs):
        if inventory_capacity == 1:
            super().__init__(
                low=0,
                high=len(OBJECT_NAMES),
                shape=(inventory_capacity,),
                name="inventory",
                **kwargs
            )
        else:
            raise NotImplementedError(
                "RLLib has a deserializing bug with the multi-discrete shape."
            )
            # space = MultiDiscrete([len(OBJECT_NAMES) for _ in range(inventory_capacity)])
            # super().__init__(low=0, high=len(OBJECT_NAMES), shape=space.shape, space=space, name="inventory", **kwargs)

    def generate(self, env, player_id, **kwargs):
        agent = env.env_agents[player_id]
        idxs = []
        for obj in agent.inventory:
            idxs.append(OBJECT_NAMES.index(obj.object_id) + 1)
        sorted_idxs = sorted(idxs)

        encoding = np.zeros(self.shape, dtype=np.uint8)
        for i, idx in enumerate(sorted_idxs):
            encoding[i] = idx

        return encoding


feature_space.register_feature("inventory", Inventory)


class CanMoveDirection(feature.Feature):
    """
    Returns a multi-hot encoding of the agent's ability to move in each direction.
    """

    def __init__(self, **kwargs):
        super().__init__(
            low=0, high=1, shape=(4,), name="can_move_direction", **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        agent = env.env_agents[player_id]
        can_move = np.zeros(self.shape, dtype=np.int32)

        # check if the agent can move in each direction by checking if the next tile in each direction
        # is overlappable (grid_obj.can_overlap())
        for i, pos in enumerate(grid_utils.adjacent_positions(*agent.pos)):
            obj = env.grid.get(*pos)

            if obj is None or obj.can_overlap(agent):
                can_move[i] = 1

        return can_move


feature_space.register_feature("can_move_direction", CanMoveDirection)


class ActionMask(feature.Feature):
    def __init__(self, env, **kwargs):
        super().__init__(
            low=0,
            high=1,
            shape=None,
            space=spaces.Box(
                env.action_spaces[0].n,
            ),
            name="action_mask",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        action_mask = env.get_action_mask(player_id)
        return action_mask


feature_space.register_feature("action_mask", ActionMask)


class AgentID(feature.Feature):
    def __init__(self, env, **kwargs):
        super().__init__(
            low=0,
            high=len(env.agent_ids) - 1,
            shape=(1,),
            name="agent_id",
            **kwargs
        )

    def generate(self, env, player_id, **kwargs):
        agent_number = (
            env.env_agents[player_id].agent_number - 1
        )  # subtract 1 so we start from 0
        return np.array([agent_number])


feature_space.register_feature("agent_id", AgentID)



"""
Testing Copied from the SSD Repo:
https://github.com/eugenevinitsky/sequential_social_dilemma_games/blob/master/tests/test_envs.py
"""

import unittest

import numpy as np
from gymnasium.spaces import Dict

from cogrid.cogrid_env import CoGridEnv
from cogrid.core.agent import Agent
from cogrid.core.actions import Actions
from cogrid.core.directions import Directions
from cogrid.constants import FIXED_GRIDS
from cogrid.constants import GridConstants
from cogrid.core import grid_object
from cogrid.envs.search_rescue import search_rescue


BASE_MAP = (
    [
        "#######",
        "#     #",
        "#     #",
        "#     #",
        "#     #",
        "#     #",
        "#######",
    ],
    np.zeros((7, 7)),
)
BASE_MAP_2 = (
    ["######", "# S  #", "#    #", "#    #", "#   S#", "######"],
    np.zeros((6, 6)),
)
SR_TEST_MAP = (
    [
        "##########",
        "#SS      #",
        "#        #",
        "#        #",
        "#       G#",
        "#        #",
        "#P     K #",
        "#XX M ##D#",
        "#GX Y #GG#",
        "##########",
    ],
    np.zeros((10, 10)),
)


class DummyAgent(Agent):
    def interact(self, char):
        return GridConstants.FreeSpace


class DummyMapEnv(CoGridEnv):
    def __init__(
        self, config: dict, test_grid_data: tuple[np.ndarray, np.ndarray]
    ):
        self.test_grid_data: tuple[np.ndarray, np.ndarray] = test_grid_data
        self.start_positions: list[tuple] | None = config.get("start_positions")
        self.start_directions: list[int] | None = config.get("start_directions")
        super().__init__(config=config)

    def _setup_agents(self) -> None:
        for agent_id in range(self.config["num_agents"]):
            agent = Agent(
                agent_id=agent_id,
                start_position=(
                    self.start_positions.pop(0)
                    if self.start_positions
                    else self.select_spawn_point()
                ),
                start_direction=(
                    self.start_directions.pop(0)
                    if self.start_directions
                    else self.np_random.choice(Directions)
                ),
            )
            self.env_agents[agent_id] = agent

    def _generate_encoded_grid_states(self, **kwargs):
        return self.test_grid_data


class TestMapEnv(unittest.TestCase):
    def tearDown(self) -> None:
        self.env = None

    def _construct_map(
        self,
        map_encoding: tuple[np.ndarray | list[str], np.ndarray],
        num_agents=1,
        start_positions=None,
        start_directions=None,
        view_size=7,
        obs=(
            "full_map_image",
            "full_map_encoding",
            "full_map_ascii",
            "fov_image",
            "fov_encoding",
        ),
        env_cls=DummyMapEnv,
    ):
        dummy_config = {
            "name": "dummyenv",
            "max_steps": 1000,
            "view_size": view_size,
            "num_agents": num_agents,
            "start_positions": start_positions,
            "start_directions": start_directions,
            "features": obs,
            "grid": {
                "layout_fn": lambda *args, **kwargs: (
                    map_encoding if type(map_encoding) == str else None
                )
            },
        }
        self.env = env_cls(dummy_config, map_encoding)
        self.env.reset()
        self.env.update_grid_agents()

    def change_agent_position(self, agent_id, new_pos, new_dir):
        agent = self.env.env_agents[agent_id]
        agent.pos = new_pos
        agent.dir = new_dir
        self.env.update_grid_agents()

    def add_agent_to_env(self, agent_id, start_position, start_direction):
        self.env.env_agents[agent_id] = Agent(
            agent_id,
            start_position=start_position,
            start_direction=start_direction,
        )
        self.env.feature_generators[agent_id] = [
            self.env._fetch_feature_generator(feature_name)
            for feature_name in self.env.config["features"]
        ]
        self.env.observation_spaces[agent_id] = Dict(
            {
                feature.name: feature.space
                for feature in self.env.feature_generators[agent_id]
            }
        )
        self.env.update_grid_agents()

    def test_step(self):
        """Check that the step method works at all for all possible actions"""
        self._construct_map(BASE_MAP_2, num_agents=1)
        aid = self.env.agent_ids[0]
        action_dim = self.env.action_space.n
        for i in range(action_dim):
            self.env.step({aid: i})

    def test_walls(self):
        """Check that the spawned map and base map have walls in the right place"""
        self._construct_map(map_encoding=BASE_MAP, num_agents=0)

        # Make sure wall objects surround the grid
        h, w = self.env.shape
        for row in range(h):
            for col in range(w):
                if row in [0, h - 1] or col in [0, w - 1]:
                    cell = self.env.grid.get(row=row, col=col)
                    assert isinstance(cell, grid_object.Wall)

        # Make sure it encodes the correct representation
        ascii_map = self.env.map_with_agents
        np.testing.assert_array_equal(ascii_map[0, :], ["#"] * w)
        np.testing.assert_array_equal(ascii_map[-1, :], ["#"] * w)
        np.testing.assert_array_equal(ascii_map[:, 0], ["#"] * h)
        np.testing.assert_array_equal(ascii_map[:, -1], ["#"] * h)

    #     env_image = self.env.get_full_render(highlight=False)
    #     wall_img = np.full_like(
    #         env_image.shape,
    #         fill_value=DEFAULT_COLORS[GridConstants.Wall],
    #     )
    #     wall_img = np.moveaxis(wall_img, -1, 0)  # need channel dim to be last for the above fill to work
    #
    #     np.testing.assert_array_equal(
    #         self.env.world_map_color[:, pad, pad:-pad],
    #         wall_img[:, pad, pad:-pad],
    #     )
    #     np.testing.assert_array_equal(
    #         self.env.world_map_color[:, -pad - 1, pad:-pad],
    #         wall_img[:, -pad - 1, pad:-pad],
    #     )
    #     np.testing.assert_array_equal(
    #         self.env.world_map_color[:, pad:-pad, pad],
    #         wall_img[:, pad:-pad, pad],
    #     )
    #     np.testing.assert_array_equal(
    #         self.env.world_map_color[:, pad:-pad, -pad - 1],
    #         wall_img[:, pad:-pad, -pad - 1],
    #     )
    #
    # def assert_logical_and_image_view(self, agent_id, expected_logical_view):
    #     obs = self.env.get_obs()
    #     image_view = obs[agent_id]["image_view"]
    #     logical_view = obs[agent_id]["ascii_view"]
    #
    #     expected_image_view = map_to_image(expected_logical_view)
    #
    #     np.testing.assert_array_equal(logical_view, expected_logical_view)
    #     np.testing.assert_array_equal(image_view, expected_image_view)
    #
    # def test_view(self):
    #     self._construct_map(BASE_MAP, num_agents=1, start_positions=[(3, 3)], view_size=2)
    #     a_id = self.env.agent_ids[0]
    #
    #     # Test that the view is empty around the agent
    #     expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "     ", "     "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Move agent one row closer to the top, check that the top row is walls
    #     self.change_agent_position(a_id, (2, 3))
    #     expected_logical_view = ascii_to_numpy(["#####", "     ", "  1  ", "     ", "     "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check that the agent sees that padded view when the row above them is out of view
    #     self.change_agent_position(a_id, (1, 3))
    #     expected_logical_view = ascii_to_numpy(["     ", "#####", "  1  ", "     ", "     "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check that the view is correct if the left wall is in view
    #     self.change_agent_position(a_id, (3, 2))
    #     expected_logical_view = ascii_to_numpy(["#    ", "#    ", "# 1  ", "#    ", "#    "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if it exceeds the left view
    #     self.change_agent_position(a_id, (3, 1))
    #     expected_logical_view = ascii_to_numpy([" #   ", " #   ", " #1  ", " #   ", " #   "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if the bottom is in view
    #     self.change_agent_position(a_id, (4, 3))
    #     expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "     ", "#####"])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if exceeding bottom view
    #     self.change_agent_position(a_id, (5, 3))
    #     expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "#####", "     "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if the right wall is in view
    #     self.change_agent_position(a_id, (3, 4))
    #     expected_logical_view = ascii_to_numpy(["    #", "    #", "  1 #", "    #", "    #"])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if exceeding the right wall
    #     self.change_agent_position(a_id, (3, 5))
    #     expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "   # ", "   # "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    #     # Check if the view is correct if in the bottom right corner
    #     self.change_agent_position(a_id, (5, 5))
    #     expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "#### ", "     "])
    #     self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # def test_agent_visibility(self):
    #     self._construct_map(BASE_MAP.copy(), 0, None, view_size=2, obs=["image_view", "other_agent_visibility"])
    #     self.add_agent_to_env("agent-0", (1, 1))
    #     self.add_agent_to_env("agent-1", (1, 3))
    #     self.add_agent_to_env("agent-2", (1, 5))
    #     obs, *_ = self.env.step({})
    #     visibility = [a_obs["other_agent_visibility"] for a_obs in obs.values()]
    #     np.testing.assert_array_equal(visibility[0], [1, 0])
    #     np.testing.assert_array_equal(visibility[1], [1, 1])
    #     np.testing.assert_array_equal(visibility[2], [0, 1])

    def test_agent_actions(self):
        a_id = 0
        self._construct_map(BASE_MAP, 1, [(2, 2)], [Directions.Up])

        self.env.step({a_id: Actions.Noop})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [2, 2]
        )
        np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")

        self.env.step({a_id: Actions.Left})
        assert self.env.grid.grid_agents[a_id].dir == Directions.Left
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 1])
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [2, 1]
        )
        np.testing.assert_array_equal(self.env.map_with_agents[2, 1], "1")

        self.env.step({a_id: Actions.Right})  # Rotate to face up
        assert self.env.grid.grid_agents[a_id].dir == Directions.Up
        self.env.step({a_id: Actions.Right})  # Rotate to face right
        assert self.env.grid.grid_agents[a_id].dir == Directions.Right
        self.env.step({a_id: Actions.Forward})  # Move back to (2, 2)
        assert self.env.grid.grid_agents[a_id].dir == Directions.Right
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [2, 2]
        )
        np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")

        self.env.step({a_id: Actions.Left})  # Rotate to face up
        assert self.env.grid.grid_agents[a_id].dir == Directions.Up
        self.env.step({a_id: Actions.Forward})
        assert self.env.grid.grid_agents[a_id].dir == Directions.Up

        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [1, 2])
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [1, 2]
        )
        np.testing.assert_array_equal(self.env.map_with_agents[1, 2], "1")

        self.env.step({a_id: Actions.Left})  # Rotate to face left
        assert self.env.grid.grid_agents[a_id].dir == Directions.Left
        self.env.step({a_id: Actions.Left})  # Rotate to face down
        assert self.env.grid.grid_agents[a_id].dir == Directions.Down
        self.env.step({a_id: Actions.Forward})
        assert self.env.grid.grid_agents[a_id].dir == Directions.Down

        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, [2, 2])
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [2, 2]
        )
        np.testing.assert_array_equal(self.env.map_with_agents[2, 2], "1")

        # if an agent tries to move through a wall they should stay in the same place
        # we check that this works correctly for both corner and non-corner edges
        self.change_agent_position(a_id, new_pos=(1, 1), new_dir=Directions.Up)
        assert self.env.env_agents[a_id].dir == Directions.Up
        assert self.env.grid.grid_agents[a_id].dir == Directions.Up
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 1))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [1, 1]
        )

        self.env.step({a_id: Actions.Left})
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 1))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [1, 1]
        )

        self.change_agent_position(
            a_id, new_pos=(4, 4), new_dir=Directions.Right
        )
        self.env.step({a_id: Actions.Forward})
        self.env.step({a_id: Actions.Right})
        self.env.step({a_id: Actions.Forward})
        self.env.step({a_id: Actions.Left})
        self.env.step({a_id: Actions.Forward})
        self.env.step({a_id: Actions.Right})

        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 5))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [5, 5]
        )

        assert self.env.grid.grid_agents[a_id].dir == Directions.Down
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 5))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [5, 5]
        )

        self.env.step({a_id: Actions.Right})
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (5, 4))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [5, 4]
        )

        self.change_agent_position(
            a_id, new_pos=(4, 5), new_dir=Directions.Right
        )
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (4, 5))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [4, 5]
        )

        self.change_agent_position(
            a_id, new_pos=(2, 1), new_dir=Directions.Left
        )
        self.env.step({a_id: Actions.Left})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (2, 1))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [2, 1]
        )

        self.change_agent_position(a_id, new_pos=(1, 2), new_dir=Directions.Up)
        self.env.step({a_id: Actions.Forward})
        np.testing.assert_array_equal(self.env.env_agents[a_id].pos, (1, 2))
        np.testing.assert_array_equal(
            self.env.grid.grid_agents[a_id].pos, [1, 2]
        )

    # def test_agent_conflict(self):
    #     self._construct_map(BASE_MAP_2, 2, [(1, 2), (4, 4)], [Directions.Up, Directions.Up])
    #
    #     # test that agents can't walk into other agents
    #     self.change_agent_position("agent-0", (3, 3))
    #     self.change_agent_position("agent-1", (3, 4))
    #     self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"]})
    #     self.env.step({"agent-1": BASE_ACTION_MAP["LEFT"]})
    #     np.testing.assert_array_equal(self.env.env_agents["agent-0"].pos, (3, 3))
    #     np.testing.assert_array_equal(self.env.env_agents["agent-1"].pos, (3, 4))
    #
    #     # test that agents can't walk through each other if they move simultaneously
    #     self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["LEFT"]})
    #     np.testing.assert_array_equal(self.env.env_agents["agent-0"].pos, (3, 3))
    #     np.testing.assert_array_equal(self.env.env_agents["agent-1"].pos, (3, 4))
    #     # also check that the map looks correct, no agent has disappeared
    #     expected_map = ascii_to_numpy(["######", "#    #", "#    #", "#  12#", "#    #", "######"])
    #
    #     np.testing.assert_array_equal(expected_map, self.env.map_with_agents)
    #
    #     # test that agents can walk into other agents if moves are de-conflicting
    #     # conflict only occurs stochastically so try it 50 times
    #     np.random.seed(1)
    #     self.change_agent_position("agent-0", (2, 4))
    #     self.change_agent_position("agent-1", (3, 4))
    #     expected_map = ascii_to_numpy(["######", "#    #", "#   1#", "#   2#", "#    #", "######"])
    #     self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["UP"]})
    #     np.testing.assert_array_equal(expected_map, self.env.map_with_agents)
    #     self.env.step({"agent-0": BASE_ACTION_MAP["LEFT"], "agent-1": BASE_ACTION_MAP["DOWN"]})
    #     self.env.step({"agent-0": BASE_ACTION_MAP["RIGHT"], "agent-1": BASE_ACTION_MAP["UP"]})
    #     np.testing.assert_array_equal(expected_map, self.env.map_with_agents)

    # def test_map_encoding(self):
    #     self._construct_map(
    #         BASE_MAP, num_agents=1, start_positions=[(3, 3)], start_directions=[Directions.Up], view_size=3
    #     )
    #     a_id = self.env.agent_ids[0]

    # # Test that the view is empty around the agent
    # expected_logical_view = ascii_to_numpy(["   ", "   ", " ^ "])
    # self.assert_encoding_and_image_view(a_id, expected_logical_view)
    #
    # # Move agent one row closer to the top, check that the top row is walls
    # self.change_agent_position(a_id, (2, 3))
    # expected_logical_view = ascii_to_numpy(["#####", "     ", "  1  ", "     ", "     "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check that the agent sees that padded view when the row above them is out of view
    # self.change_agent_position(a_id, (1, 3))
    # expected_logical_view = ascii_to_numpy(["     ", "#####", "  1  ", "     ", "     "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check that the view is correct if the left wall is in view
    # self.change_agent_position(a_id, (3, 2))
    # expected_logical_view = ascii_to_numpy(["#    ", "#    ", "# 1  ", "#    ", "#    "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if it exceeds the left view
    # self.change_agent_position(a_id, (3, 1))
    # expected_logical_view = ascii_to_numpy([" #   ", " #   ", " #1  ", " #   ", " #   "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if the bottom is in view
    # self.change_agent_position(a_id, (4, 3))
    # expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "     ", "#####"])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if exceeding bottom view
    # self.change_agent_position(a_id, (5, 3))
    # expected_logical_view = ascii_to_numpy(["     ", "     ", "  1  ", "#####", "     "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if the right wall is in view
    # self.change_agent_position(a_id, (3, 4))
    # expected_logical_view = ascii_to_numpy(["    #", "    #", "  1 #", "    #", "    #"])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if exceeding the right wall
    # self.change_agent_position(a_id, (3, 5))
    # expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "   # ", "   # "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)
    #
    # # Check if the view is correct if in the bottom right corner
    # self.change_agent_position(a_id, (5, 5))
    # expected_logical_view = ascii_to_numpy(["   # ", "   # ", "  1# ", "#### ", "     "])
    # self.assert_logical_and_image_view(a_id, expected_logical_view)

    def test_interact_sequence(self):
        self._construct_map(
            FIXED_GRIDS["small_test_map"],
            num_agents=1,
            start_positions=[(1, 1)],
            start_directions=[Directions.Down],
            view_size=7,
        )

        env_config = {
            "name": "search_rescue",
            "num_agents": 1,
            "features": [
                "full_map_encoding",
                "fov_encoding",
            ],
            # TODO(chase): register this layout
            "grid": {
                "layout": "small_test_map",
            },
            "roles": False,
            "agent_view_size": 7,
            "max_steps": 1000,
        }

        self.env = search_rescue.SearchRescueEnv(env_config)
        agent_id = self.env.agent_ids[0]
        self.change_agent_position(agent_id, (1, 1), Directions.Down)

        # Starts in top left facing down, move 4 units down to the Pickaxe
        self.env.step({agent_id: Actions.Forward})
        self.env.step({agent_id: Actions.Forward})
        self.env.step({agent_id: Actions.Forward})
        obs, *_ = self.env.step({agent_id: Actions.Forward})

        # Pick up the pickaxe and move towards the rubble
        obs, *_ = self.env.step({agent_id: Actions.Pickup})
        self.env.step({agent_id: Actions.Forward})


if __name__ == "__main__":
    unittest.main()



import dataclasses

import numpy as np

from cogrid import constants
from cogrid.core import grid_utils
from cogrid.core import grid_object


def generate_sr_grid(
    shape=None,
    load=None,
    num_green=4,
    num_yellow=3,
    num_red=2,
    num_agents=2,
    populate_rubble=True,
    np_random=None,
):
    if np_random is None:
        np_random = np.random.RandomState(seed=42)

    if load is not None:
        return getattr(FixedGrids, load)

    # First channel is characters, second channel is state
    grid = np.full((*shape, 2), fill_value=constants.GridConstants.FreeSpace)
    grid[:, :, 1] = 0

    # Fill outside border with walls
    grid[0, :] = grid_object.Wall.char
    grid[-1, :] = grid_object.Wall.char
    grid[:, 0] = grid_object.Wall.char
    grid[:, -1] = grid_object.Wall.char

    # Get free space indices
    free_spaces = list(np.argwhere(grid[:, :, 0] == constants.GridConstants.FreeSpace))

    objs_to_place = num_green + num_yellow + num_red + num_agents
    assert (
        len(free_spaces) >= objs_to_place
    ), "Not enough free spaces for specified number of objects!"

    np_random.shuffle(free_spaces)

    for _ in range(num_green):
        r, c = free_spaces.pop()
        grid[r, c, 0] = "G"

    for _ in range(num_yellow):
        r, c = free_spaces.pop()
        grid[r, c, 0] = "Y"

    for _ in range(num_red):
        r, c = free_spaces.pop()
        grid[r, c, 0] = "R"

    # add spawns
    for _ in range(num_agents):
        r, c = free_spaces.pop()
        grid[r, c, 0] = "S"

    if populate_rubble:
        # populate rubble around yellow and red
        for r, c in np.argwhere(grid == "Y"):
            grid[:, :, 0] = surround_by_rubble(grid[:, :, 0], r, c)
        for r, c in np.argwhere(grid == "R"):
            grid[:, :, 0] = surround_by_rubble(grid[:, :, 0], r, c)

    return grid


def surround_by_rubble(grid, row, col):
    for r, c in grid_utils.adjacent_positions(row, col):
        if grid[r, c] == constants.GridConstants.FreeSpace:
            grid[r, c] = constants.GridConstants.Rubble
    return grid


@dataclasses.dataclass
class FixedGrids:
    m3minimap = [
        "#############",
        "#S  S#      #",
        "#  X #      #",
        "# XR #XYXG  #",
        "#    # X  XY#",
        "#   X#     X#",
        "#  XY#      #",
        "#    #  RX  #",
        "##X###  X   #",
        "#    #     X#",
        "#    #G   XR#",
        "#  X #     X#",
        "# XYX#    # #",
        "#    ###X# ##",
        "#G   #    G #",
        "#    X     G#",
        "#    #X     #",
        "#    #YXX # #",
        "# X  #X RX ##",
        "# RX #  #####",
        "#   #########",
        "#############",
    ]

    m3minimap_reduced = [
        "#############",
        "#S  S       #",
        "#  X   X    #",
        "# XRX XYXG  #",
        "#  XG  X  XY#",
        "#  XYX      #",
        "#   X     XX#",
        "#XX       XX#",
        "#RX  X    XR#",
        "#############",
    ]



"""
Grid representation derived from Minigrid:
https://github.com/Farama-Foundation/Minigrid/minigrid/core/grid.py
"""

from __future__ import annotations

from typing import Any, Callable
from copy import deepcopy

import numpy as np

from cogrid.core.grid_object import GridObj, Wall, object_to_idx, GridAgent
from cogrid.core import grid_object
from cogrid.core.constants import CoreConstants
from cogrid.constants import GridConstants
from cogrid.visualization.rendering import (
    downsample,
    fill_coords,
    highlight_img,
    point_in_rect,
)

CHANNEL_FIRST = False


def get_grid_agent_at_position(
    grid: Grid, position: tuple[int, int] | np.ndarray
) -> GridAgent | None:
    """Return the GridAgent at a given position, if any.

    :param grid: The grid to search.
    :type grid: Grid
    :param position: The position to search for.
    :type position: tuple[int, int] | np.ndarray
    :return: The GridAgent at the given position, if any.
    :rtype: GridAgent | None
    """
    for grid_agent in grid.grid_agents.values():
        # assert agent.pos is not None, "Agent pos should never be None."
        if np.array_equal(grid_agent.pos, position):
            return grid_agent
    return None


class Grid:
    """The Grid class is a 2D HxW grid of GridObjs.

    :param height: Height of the grid.
    :type height: int
    :param width: Width of the grid.
    :type width: int
    """

    tile_cache: dict[tuple[Any], Any] = {}

    def __init__(self, height: int, width: int):
        """Generator method for Grid class."""
        assert height >= 3 and width >= 3, "Both dimensions must be >= 3."

        self.height: int = height
        self.width: int = width

        self.grid: list[GridObj | None] = [None] * (height * width)
        self.grid_agents: dict[str, GridAgent] = {}

    def __contains__(self, item: Any) -> bool:
        """Check if an item is in the grid.

        :param item: The item to check for.
        :type item: Any
        :return: Whether the item is in the grid.
        :rtype: bool
        """
        # TODO(chase): Add the ability to pass an encoding.
        if isinstance(item, GridObj):
            for grid_obj in self.grid:
                if grid_obj is item:
                    return True
        return False

    def __eq__(self, other_grid: Grid) -> bool:
        """Check if two grids are equal.

        :param other_grid: The other grid to compare.
        :type other_grid: Grid
        :return: Whether the two grids are equal.
        :rtype: bool
        """
        g1 = self.encode()
        g2 = other_grid.encode()
        return np.array_equal(g1, g2)

    def __ne__(self, other: Grid) -> bool:
        """Check if two grids are not equal.

        :param other: The other grid to compare.
        :type other: Grid
        :return: Whether the two grids are not equal.
        :rtype: bool
        """
        return not self == other

    def copy(self) -> Grid:
        """Return a deep copy of the grid.

        :return: A deep copy of the grid.
        :rtype: Grid
        """
        return deepcopy(self)

    def set(self, row: int, col: int, v: GridObj | None) -> None:
        """Set a GridObj at a given position in the grid.

        :param row: The row index.
        :type row: int
        :param col: The column index.
        :type col: int
        :param v: The GridObj (or None) to set.
        :type v: GridObj | None
        """
        assert (
            0 <= col < self.width
        ), f"column index {col} outside of grid of width {self.width}"
        assert (
            0 <= row < self.height
        ), f"row index {row} outside of grid of height {self.height}"
        self.grid[row * self.width + col] = v

    def get(self, row: int, col: int) -> GridObj | None:
        """Get the GridObj at a given position in the grid.

        :param row: The row index.
        :type row: int
        :param col: The column index.
        :type col: int
        :return: The GridObj at the given position, if any.
        :rtype: GridObj | None
        """
        assert (
            0 <= col < self.width
        ), f"column index {col} outside of grid of width {self.width}"
        assert (
            0 <= row < self.height
        ), f"row index {row} outside of grid of height {self.height}"
        assert self.grid is not None
        return self.grid[row * self.width + col]

    def tick(self):
        """Tick the grid, which calls the tick() method of all GridObjs.
        This is useful for any GridOjbs with time-dependent behavior.
        """
        for grid_obj in self.grid:
            if grid_obj is not None:
                grid_obj.tick()

    def horz_wall(
        self,
        col: int,
        row: int,
        length: int | None = None,
        obj_type: GridObj = Wall,
    ) -> None:
        """Create a horizontal wall of obj_type in the grid.

        :param col: The column index.
        :type col: int
        :param row: The row index.
        :type row: int
        :param length: The length of the wall, defaults to maximum.
        :type length: int | None, optional
        :param obj_type: GridObj to construct a wall from, defaults to Wall.
        :type obj_type: Callable[[], GridObj], optional
        """
        if length is None:
            length = self.width - col
        for i in range(length):
            self.set(row=row, col=col + i, v=obj_type())

    def vert_wall(
        self,
        col: int,
        row: int,
        length: int | None = None,
        obj_type: GridObj = Wall,
    ) -> None:
        """Create a vertical wall of obj_type in the grid.

        :param col: The column index.
        :type col: int
        :param row: The row index.
        :type row: int
        :param length: The length of the wall, defaults to maximum.
        :type length: int | None, optional
        :param obj_type: GridObj to construct a wall from, defaults to Wall
        :type obj_type: Callable[[], GridObj], optional
        """
        if length is None:
            length = self.height - row
        for j in range(length):
            self.set(row=row + j, col=col, v=obj_type())

    def wall_rect(
        self, col: int, row: int, w: int, h: int, grid_obj: GridObj = Wall
    ):
        """Create a rectangle of walls in the grid.

        :param col: Initial column index.
        :type col: int
        :param row: Initial row index.
        :type row: int
        :param w: Width of the rectangle.
        :type w: int
        :param h: Height of the rectangle.
        :type h: int
        :param grid_obj: GridObj to construct a wall from, defaults to Wall.
        :type grid_obj: GridObj, optional
        """
        self.horz_wall(row=row, col=col, length=w, obj_type=grid_obj)
        self.horz_wall(row=row, col=col + h - 1, length=w, obj_type=grid_obj)
        self.vert_wall(row=row, col=col, length=h, obj_type=grid_obj)
        self.vert_wall(row=row + w - 1, col=col, length=h, obj_type=grid_obj)

    def rotate_left(self) -> Grid:
        """Rotate the grid to the left (counter-clockwise)

        :return: The rotated grid.
        :rtype: Grid
        """
        grid = Grid(width=self.height, height=self.width)

        for col in range(self.width):
            for row in range(self.height):
                v = self.get(row=row, col=col)
                if v:
                    v.rotate_left()

                new_row = grid.height - 1 - col
                new_col = row
                grid.set(row=new_row, col=new_col, v=v)

                agent = get_grid_agent_at_position(self, (row, col))
                if agent:
                    rotated_agent = deepcopy(agent)
                    rotated_agent.rotate_left()
                    rotated_agent.pos = (new_row, new_col)
                    grid.grid_agents[rotated_agent.agent_id] = rotated_agent

        return grid

    def slice(self, topX: int, topY: int, width: int, height: int) -> Grid:
        """Get a subset of the grid

        :param topX: The top x-coordinate of the slice.
        :type topX: int
        :param topY: The top y-coordinate of the slice.
        :type topY: int
        :param width: Width of the slice.
        :type width: int
        :param height: Height of the slice.
        :type height: int
        :return: The sliced grid.
        :rtype: Grid
        """
        grid = Grid(height=height, width=width)
        for row in range(height):
            for col in range(width):
                x = topX + col
                y = topY + row
                if 0 <= x < self.width and 0 <= y < self.height:
                    v = self.get(row=y, col=x)

                    agent = get_grid_agent_at_position(
                        grid=self, position=(y, x)
                    )
                    if agent is not None:
                        grid_slice_agent = deepcopy(agent)
                        grid_slice_agent.pos = (row, col)
                        grid.grid_agents[grid_slice_agent.agent_id] = (
                            grid_slice_agent
                        )
                else:
                    v = Wall()

                grid.set(row=row, col=col, v=v)

        return grid

    def render_tile(
        self,
        obj: GridObj | None,
        highlight: bool = False,
        position: tuple[int, int] | None = None,
        tile_size: int = CoreConstants.TilePixels,
        subdivs: int = 3,
    ) -> np.ndarray:
        """Render a tile and cache the result.

        :param obj: The GridObj to render on the tile, defaults to None.
        :type obj: GridObj | None
        :param highlight: If the cell should be highlighted, defaults to False
        :type highlight: bool, optional
        :param position: Position of the GridObj (useful for retrieving GridAgent), defaults to None
        :type position: tuple[int, int] | None, optional
        :param tile_size: Rendered size of the tile, defaults to CoreConstants.TilePixels
        :type tile_size: int, optional
        :param subdivs: Number of sub-divisions of the tile, defaults to 3
        :type subdivs: int, optional
        :return: An RGB image of the rendered tile.
        :rtype: np.ndarray
        """
        grid_agent = get_grid_agent_at_position(grid=self, position=position)
        agent_dir = grid_agent.dir if grid_agent else None
        agent_color = grid_agent.agent_id if grid_agent else None
        agent_inventory_names = (
            tuple([obj.object_id for obj in grid_agent.inventory])
            if grid_agent
            else None
        )
        key: tuple[Any, ...] = (
            (agent_dir, agent_color, agent_inventory_names),
            highlight,
            tile_size,
        )
        key = obj.encode() + key if obj else key

        if key in self.__class__.tile_cache:
            return self.__class__.tile_cache[key]

        tile_img = np.zeros(
            shape=(tile_size * subdivs, tile_size * subdivs, 3), dtype=np.uint8
        )

        # Draw grid lines (separating each tile)
        fill_coords(tile_img, point_in_rect(0, 0.031, 0, 1), (100, 100, 100))
        fill_coords(tile_img, point_in_rect(0, 1, 0, 0.031), (100, 100, 100))

        # Render the object itself onto the tile
        if obj is not None:
            obj.render(tile_img)

        if grid_agent is not None:
            grid_agent.render(tile_img)

        if highlight:
            highlight_img(tile_img)

        tile_img = downsample(tile_img, subdivs)

        # Cache the rendered tile
        self.__class__.tile_cache[key] = tile_img

        return tile_img

    def render(
        self,
        tile_size: int,
        highlight_mask: np.ndarray | None = None,
    ) -> np.ndarray:
        """Render a tile and cache the result.

        :param tile_size: The size of the tile (pixels).
        :type tile_size: int
        :param highlight_mask: Array mask to indicate which cells are highlighted, defaults to None
        :type highlight_mask: np.ndarray | None, optional
        :return: An RGB image of the rendered grid.
        :rtype: np.ndarray
        """

        # TODO(chase): Make efficiency improvements here. One option is to use
        # "dirty tiles" so we only render the tiles that have changed.

        if highlight_mask is None:
            highlight_mask = np.zeros(
                shape=(self.width, self.height), dtype=bool
            )

        # Compute total size
        width_px = self.width * tile_size
        height_px = self.height * tile_size

        img = np.zeros(shape=(height_px, width_px, 3), dtype=np.uint8)

        # render the grid
        assert highlight_mask is not None
        for row in range(self.height):
            for col in range(self.width):
                cell = self.get(row=row, col=col)

                tile_img = self.render_tile(
                    cell,
                    highlight=highlight_mask.T[row, col],
                    position=(row, col),
                    tile_size=tile_size,
                )
                ymin = row * tile_size
                ymax = (row + 1) * tile_size
                xmin = col * tile_size
                xmax = (col + 1) * tile_size

                img[ymin:ymax, xmin:xmax, :] = tile_img

        return img

    def encode(
        self, vis_mask: np.ndarray | None = None, encode_char=False
    ) -> np.ndarray:
        """Produce an ASCII/int representation of the grid."""
        if vis_mask is None:
            vis_mask = np.ones((self.height, self.width), dtype=bool)

        array = np.empty((self.height, self.width, 3), dtype=object)

        assert vis_mask is not None
        for col in range(self.width):
            for row in range(self.height):
                if not vis_mask[row, col]:
                    continue

                v = self.get(row=row, col=col)
                if v is None:
                    encoding = (
                        (
                            GridConstants.FreeSpace
                            if encode_char
                            else object_to_idx(None)
                        ),
                        0,
                        0,
                    )
                else:
                    encoding = v.encode(encode_char=encode_char)

                array[row, col] = encoding

        for grid_agent in self.grid_agents.values():
            row, col = grid_agent.pos
            array[row, col] = grid_agent.encode(encode_char=encode_char)

        if not encode_char:
            array = array.astype(np.int8)

        return array

    @staticmethod
    def decode(
        array: np.ndarray, scope: str = "global"
    ) -> tuple[Grid, np.ndarray]:
        """Decode ASCII encoding back into a Grid"""

        channels, height, width = array.shape
        assert channels == 2

        vis_mask = np.ones(shape=(height, width), dtype=bool)

        grid = Grid(height=height, width=width)
        agent_count = 0
        for col in range(width):
            for row in range(height):
                char, state = array[:, row, col]
                state = int(float(state))
                v = GridObj.decode(char, state, scope=scope)
                if v:
                    v.pos = v.init_pos = (row, col)
                    vis_mask[row, col] = v.visible()

                if isinstance(v, GridAgent):
                    grid.grid_agents[agent_count] = v
                    agent_count += 1
                else:
                    grid.set(row=row, col=col, v=v)

        return grid, vis_mask

    def get_obj_count(self, grid_obj: GridObj | None) -> int:
        """Get the number of a particular object that exists in the grid.

        :param grid_obj: The GridObj to count.
        :type grid_obj: GridObj | None
        :return: The number of the GridObj in the grid.
        :rtype: int
        """
        count = 0
        for obj in self.grid:
            if isinstance(obj, grid_obj):
                count += 1
        return count

    def process_vis(self, agent_pos: tuple[int, int]) -> np.ndarray:
        mask = np.zeros(shape=(self.width, self.height), dtype=bool)

        mask[agent_pos[1], agent_pos[0]] = True

        for j in reversed(range(0, self.height)):
            for i in range(0, self.width - 1):
                if not mask[i, j]:
                    continue

                cell = self.get(j, i)
                if cell and not cell.see_behind():
                    continue

                mask[i + 1, j] = True
                if j > 0:
                    mask[i + 1, j - 1] = True
                    mask[i, j - 1] = True

            for i in reversed(range(1, self.width)):
                if not mask[i, j]:
                    continue

                cell = self.get(j, i)
                if cell and not cell.see_behind():
                    continue

                mask[i - 1, j] = True
                if j > 0:
                    mask[i - 1, j - 1] = True
                    mask[i, j - 1] = True

        for j in range(0, self.height):
            for i in range(0, self.width):
                if not mask[i, j]:
                    self.set(j, i, None)

        mask = np.transpose(mask, axes=(1, 0))
        return mask



import collections
from itertools import combinations
import copy

import numpy as np


try:
    import pygame
except ImportError:
    pygame = None


from gymnasium import spaces
import pettingzoo
from cogrid import constants
from cogrid.core import actions as grid_actions
from cogrid.core.constants import CoreConstants
from cogrid.core import directions
from cogrid.core import grid
from cogrid.core import grid_object
from cogrid.core import grid_utils
from cogrid.feature_space import feature_space
from cogrid.core import reward
from cogrid.core import typing
from cogrid.core import agent
from cogrid.core import directions
from cogrid.core import reward

from cogrid.core import layouts


RNG = RandomNumberGenerator = np.random.Generator


class CoGridEnv(pettingzoo.ParallelEnv):
    """CoGridEnv is the base environment class for a multi-agent grid-world environment.

    This class inherits from the ``pettingzoo.ParallelEnv`` class and implements the necessary methods
    for a parallel environment.

    :param config: Configuration dictionary for the environment.
    :type config: dict
    :param render_mode: Rendering method for local visualization, defaults to None
    :type render_mode: str | None, optional
    :param agent_class: Agent class for the environment if using a custom Agent, defaults to None
    :type agent_class: agent.Agent | None, optional
    :raises ValueError: ValueError if an invalid or None action set string is provided.
    """

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 35,
        "screen_size": 480,
        "render_message": "",
        "agent_pov": None,
        "highlight": False,
        "see_through_walls": True,
    }

    def __init__(
        self,
        config: dict,
        render_mode: str | None = None,
        agent_class: agent.Agent | None = None,
        **kwargs,
    ):
        """_summary_

        :param config: _description_
        :type config: dict
        :param render_mode: _description_, defaults to None
        :type render_mode: str | None, optional
        :param agent_class: _description_, defaults to None
        :type agent_class: agent.Agent | None, optional
        :raises ValueError: _description_
        """
        super(CoGridEnv, self).__init__()
        self._np_random: np.random.Generator | None = None  # set in reset()

        # TODO(chase): Move PyGame/rendering logic outside of this class.
        self.clock = None
        self.render_size = None
        self.config = config
        self.render_mode = render_mode
        self.render_message = (
            kwargs.get("render_message") or self.metadata["render_message"]
        )
        self.tile_size = CoreConstants.TilePixels
        self.screen_size = (
            kwargs.get("screen_size") or self.metadata["screen_size"]
        )
        self.window = None
        self.name = config["name"]
        self.cumulative_score = 0

        self.max_steps = config["max_steps"]
        self.visualizer = None
        self.roles = self.config.get("roles", True)
        self.agent_class = agent_class or agent.Agent
        self.t = 0

        # grid data is set by _gen_grid()
        self.scope: str = config.get("scope", "global")
        self.grid: grid.Grid | None = None
        self.spawn_points: list = []
        self.current_layout_id: str | None = None
        self._gen_grid()
        self.shape = (self.grid.height, self.grid.width)

        self.agent_view_size = self.config.get("agent_view_size", 7)

        self.possible_agents = [i for i in range(config["num_agents"])]
        self.agents = copy.copy(self.possible_agents)
        self._agent_ids: set[typing.AgentID] = set(self.agents)
        self.env_agents: dict[typing.AgentID, agent.Agent] = {
            i: None for i in self.agents
        }  # will contain: {'agent_id': agent}

        # Establish reward function through reward modules
        reward_names = config.get("rewards", [])
        self.rewards = []
        for reward_name in reward_names:
            self.rewards.append(
                reward.make_reward(reward_name, agent_ids=self._agent_ids)
            )

        # The reward at each timestep is the sum of the rewards from each reward module
        # but can also be added to via environment hooks.
        self.per_agent_reward: dict[typing.AgentID, float] = (
            self.get_empty_reward_dict()
        )
        self.per_component_reward: dict[str, dict[typing.AgentID, float]] = {}
        self.reward_this_step = self.get_empty_reward_dict()

        # Action space describes the set of actions available to agents.
        action_str = config.get("action_set")
        if action_str == "rotation_actions":
            self.action_set = grid_actions.ActionSets.RotationActions
        elif action_str == "cardinal_actions":
            self.action_set = grid_actions.ActionSets.CardinalActions
        else:
            raise ValueError(
                f"Invalid or None action set string: {action_str}."
            )

        # Set the action space for the gym environment
        self.action_spaces = {
            a_id: spaces.Discrete(len(self.action_set))
            for a_id in self.agent_ids
        }

        # Establish the observation space. This provides the general form of what an agent's observation
        # looks like so that they can be properly initialized.
        features = config.get("features", [])
        if isinstance(features, list):
            features = {agent_id: features for agent_id in self.agent_ids}
        elif isinstance(features, str):
            features = {agent_id: [features] for agent_id in self.agent_ids}
        else:
            assert isinstance(features, dict) and set(features.keys()) == set(
                self.agent_ids
            ), (
                "Must pass a feature dictionary keyed by agent IDs, "
                "a list of features (universal for all agents), or the name "
                "of a single feature."
            )

        self.feature_spaces = {
            a_id: feature_space.FeatureSpace(
                feature_names=features[a_id], env=self, agent_id=a_id
            )
            for a_id in self.agent_ids
        }

        self.observation_spaces = {
            a_id: self.feature_spaces[a_id].observation_space
            for a_id in self.agent_ids
        }

        self.prev_actions = None

    def _gen_grid(self) -> None:
        """Generates the grid for the environment.

        This method generates the grid for the environment by calling the ``_generate_encoded_grid_states`` method,
        converting the grid to the correct numpy format, finding the spawn points, and encoding the grid and states.
        The resulting grid is then decoded and stored in the ``self.grid`` attribute.
        """
        self.spawn_points = []
        encoded_grid, states = self._generate_encoded_grid_states()

        # If the grid is a list of strings this will turn it into the correct np format
        np_grid = grid_utils.ascii_to_numpy(encoded_grid)
        spawn_points = np.where(np_grid == constants.GridConstants.Spawn)
        np_grid[spawn_points] = constants.GridConstants.FreeSpace
        self.spawn_points = list(zip(*spawn_points))
        grid_encoding = np.stack([np_grid, states], axis=0)
        self.grid, _ = grid.Grid.decode(grid_encoding, scope=self.scope)

    def _generate_encoded_grid_states(self) -> tuple[np.ndarray, np.ndarray]:
        """Generates a grid encoding from the configuration.

        :return: A tuple containing the encoded grid and state arrays.
        :rtype: tuple[np.ndarray, np.ndarray]
        """
        grid_config: dict = self.config.get("grid", {})
        layout_name = grid_config.get("layout", None)
        layout_fn = grid_config.get("layout_fn", None)

        if layout_name is None and layout_fn is None:
            raise ValueError(
                "Must provide either a `layout` name or layout-generating function in config['grid']"
            )
        elif layout_name is not None:
            layout, state_encoding = layouts.get_layout(layout_name)
            self.current_layout_id = layout_name
        elif layout_fn is not None:
            layout_name, layout, state_encoding = layout_fn(**grid_config)
            self.current_layout_id = layout_name

        return layout, state_encoding

    @staticmethod
    def _set_np_random(
        seed: int | None = None,
    ) -> tuple[RandomNumberGenerator, int]:
        """Set the numpy random number generator. This is copied from
        the

        :param seed: _description_, defaults to None
        :type seed: int | None, optional
        :raises ValueError: Invalid seed value.
        :return: Random number generator and seed.
        :rtype: tuple[RandomNumberGenerator, int]
        """
        if seed is not None and not (isinstance(seed, int) and 0 <= seed):
            if isinstance(seed, int) is False:
                raise ValueError(
                    f"Seed must be a python integer, actual type: {type(seed)}"
                )
            else:
                raise ValueError(
                    f"Seed must be greater or equal to zero, actual value: {seed}"
                )

        seed_seq = np.random.SeedSequence(seed)
        np_seed = seed_seq.entropy
        rng = RandomNumberGenerator(np.random.PCG64(seed_seq))
        return rng, np_seed

    @property
    def np_random(self) -> np.random.Generator:
        """Get the numpy random number generator.

        :return: The numpy random number generator.
        :rtype: np.random.Generator
        """
        if self._np_random is None:
            self._np_random, _ = self._set_np_random()

        return self._np_random

    def observation_space(self, agent: typing.AgentID) -> spaces.Space:
        """Takes in agent and returns the observation space for that agent.

        :param agent: The agent ID.
        :type agent: typing.AgentID
        :return: The observation space for the agent.
        :rtype: spaces.Space
        """
        return self.observation_spaces[agent]

    def action_space(self, agent: typing.AgentID) -> spaces.Space:
        """Takes in agent and returns the action space for that agent.

        :param agent: The agent ID.
        :type agent: typing.AgentID
        :return: The action space for the agent.
        :rtype: spaces.Space
        """
        return self.action_spaces[agent]

    def reset(
        self,
        *,
        seed: int | None = 42,
        options: dict[str, typing.Any] | None = None,
    ) -> tuple[dict[typing.AgentID, typing.ObsType], dict[str, typing.Any]]:
        """Reset the environement and return the initial observations. Must be implemented for each environment.

        :param seed: NumPy random seed, defaults to 42
        :type seed: int | None, optional
        :param options: Environment reset options, defaults to None
        :type options: dict[str, typing.Any] | None, optional
        :return: Tuple of observations and info.
        :rtype: tuple[dict[typing.AgentID, typing.ObsType], dict[str, typing.Any]]
        """
        if seed is not None:
            self._np_random, _ = self._set_np_random(seed=seed)

        self.agents = copy.copy(self.possible_agents)

        self._gen_grid()

        # Clear out past agents and re-initialize
        self.env_agents = {}
        self.setup_agents()

        # Initialize previous actions as no-op
        self.prev_actions = {
            a_id: grid_actions.Actions.Noop for a_id in self.agent_ids
        }

        self.t = 0

        self.on_reset()

        self.update_grid_agents()

        self.cumulative_score = 0

        self.render()

        obs = self.get_obs()

        self.per_agent_reward = self.get_empty_reward_dict()
        self.per_component_reward = {}

        return obs, {agent_id: {} for agent_id in self.agent_ids}

    def _action_idx_to_str(
        self, actions: dict[typing.AgentID, int | str]
    ) -> dict[typing.AgentID, str]:
        """Convert the action from index to string representation


        :param actions: Dictionary of agent IDs and actions.
        :type actions: dict[str, int  |  str]
        :return: _description_
        :rtype: dict[str, str]
        """
        str_actions = {
            a_id: (
                self.action_set[action]
                if not isinstance(action, str)
                else action
            )
            for a_id, action in actions.items()
        }

        return str_actions

    def step(self, actions: dict[typing.AgentID, typing.ActionType]) -> tuple[
        dict[typing.AgentID, typing.ObsType],
        dict[typing.AgentID, float],
        dict[typing.AgentID, bool],
        dict[typing.AgentID, bool],
        dict[typing.AgentID, dict[typing.Any, typing.Any]],
    ]:
        """Transitition the environment forward by one step, given the actions of the agents.

        :param actions: Dictionary of agent IDs and actions.
        :type actions: dict
        :return: Tuple of observations, rewards, terminateds, truncateds, and infos.
        :rtype: tuple[ dict[typing.AgentID, typing.ObsType], dict[typing.AgentID, float], dict[typing.AgentID, bool], dict[typing.AgentID, bool], dict[typing.AgentID, dict[typing.Any, typing.Any]], ]
        """
        self.t += 1

        # Reset the rewards for this step
        self.per_agent_reward = self.get_empty_reward_dict()
        self.per_component_reward = {}

        # Track the previous state so that we have the delta for computing rewards
        self.prev_grid = copy.deepcopy(self.grid)

        # Update attributes of the grid objects that are timestep dependent
        self.grid.tick()

        # Convert the integer actions to strings (helpful for debugging!)
        actions = self._action_idx_to_str(actions)

        # Agents who are moving have their positions updated (and conflicts resolved)
        self.move_agents(actions)

        # Given any new position(s), agents interact with the environment
        self.interact(actions)

        # Updates the GridAgent objects to reflect new positions/interactions
        self.update_grid_agents()

        # Store the actions taken by the agents
        self.prev_actions = actions.copy()

        # Setup the return values
        observations = self.get_obs()
        self.compute_rewards()
        infos = self.get_infos(**self.per_component_reward)
        terminateds, truncateds = self.get_terminateds_truncateds()

        self.render()

        self.cumulative_score += sum([*self.per_agent_reward.values()])

        # Custom hook if a subclass wants to make any updates
        self.on_step()

        return (
            observations,
            self.per_agent_reward,
            terminateds,
            truncateds,
            infos,
        )

    def update_grid_agents(self) -> None:
        """Update the grid agents to reflect the current state of each Agent."""
        self.grid.grid_agents = {
            a_id: grid_object.GridAgent(agent)
            for a_id, agent in self.env_agents.items()
        }

    def setup_agents(self):
        self._setup_agents()

    def _setup_agents(self):
        for agent_id in range(self.config["num_agents"]):
            agent = self.agent_class(
                agent_id=agent_id,
                start_position=self.select_spawn_point(),
                start_direction=self.np_random.choice(directions.Directions),
            )
            self.env_agents[agent_id] = agent

    def move_agents(
        self, actions: dict[typing.AgentID, typing.ActionType]
    ) -> None:
        """Move agents to new positions based on the actions they take.

        :param actions: A dictionary of agent IDs and the actions they are taking.
        :type actions: dict[typing.AgentID, typing.ActionType]
        """
        # All terminated agents or those we don't have an action for will stay in the same position
        new_positions = {
            a_id: agent.pos
            for a_id, agent in self.env_agents.items()
            if agent.terminated or a_id not in actions
        }

        # All agents that are taking an action will be moved to a new position
        agents_to_move = [
            a_id
            for a_id in actions.keys()
            if not self.env_agents[a_id].terminated
        ]

        # If we're using cardinal actions, change the agent direction if they aren't
        # already facing the desired dir if they are, then they move in that direction.
        if self.action_set == grid_actions.ActionSets.CardinalActions:
            move_action_to_dir = {
                grid_actions.Actions.MoveRight: directions.Directions.Right,
                grid_actions.Actions.MoveLeft: directions.Directions.Left,
                grid_actions.Actions.MoveUp: directions.Directions.Up,
                grid_actions.Actions.MoveDown: directions.Directions.Down,
            }
            for a_id, action in actions.items():
                if action in move_action_to_dir.keys():
                    agent = self.env_agents[a_id]
                    desired_direction = move_action_to_dir[action]

                    # rotate to the desired direction and move that way
                    agent.dir = desired_direction
                    actions[a_id] = grid_actions.Actions.Forward

        # Determine the position each agent is attempting to move to
        attempted_positions = {}
        for a_id, action in actions.items():
            attempted_positions[a_id] = self.determine_attempted_pos(
                a_id, action
            )

        # First, give priority to agents staying in the same position
        for a_id, attemped_pos in attempted_positions.items():
            agent = self.env_agents[a_id]
            if np.array_equal(attemped_pos, agent.pos):
                new_positions[a_id] = agent.pos
                if a_id in agents_to_move:
                    agents_to_move.remove(a_id)

        # Randomize agent priority and move agents to new positions
        self.np_random.shuffle(agents_to_move)
        for a_id in agents_to_move:
            agent = self.env_agents[a_id]
            attempted_pos = attempted_positions[a_id]

            # If another agent is already moving to the desired position,
            # keep them at the current position
            if tuple(attempted_pos) in [
                tuple(npos) for npos in new_positions.values()
            ]:
                new_positions[a_id] = agent.pos
            else:
                new_positions[a_id] = attempted_pos

        # Make sure no two agents moved through each other
        for a_id1, a_id2 in combinations(self.agent_ids, r=2):
            agent1, agent2 = self.env_agents[a_id1], self.env_agents[a_id2]
            if np.array_equal(
                new_positions[a_id1], agent2.pos
            ) and np.array_equal(new_positions[a_id2], agent1.pos):
                new_positions[a_id1] = agent1.pos
                new_positions[a_id2] = agent2.pos

        # assign the new positions and store the agent position
        for a_id, agent in self.env_agents.items():
            agent.pos = new_positions[a_id]
            self.on_move(a_id)

        assert len(self.agent_pos) == len(
            set(self.agent_pos)
        ), "Agents do not have unique positions!"

    def determine_attempted_pos(
        self, agent_id: typing.AgentID, action: typing.ActionType
    ) -> tuple[int, int]:
        """Determine the position an agent is attempting to move to.

        :param agent_id: The ID of the agent attempting to move.
        :type agent_id: typing.AgentID
        :param action: The action the agent is attempting to take.
        :type action: typing.ActionType
        :return: The position the agent is attempting to move to.
        :rtype: tuple[int, int]
        """
        agent = self.env_agents[agent_id]
        fwd_pos = agent.front_pos
        fwd_cell = self.grid.get(*fwd_pos)

        if action == grid_actions.Actions.Forward:
            if fwd_cell is None or fwd_cell.can_overlap(agent=agent):
                return fwd_pos

        return agent.pos

    def can_toggle(self, agent_id: typing.AgentID) -> bool:
        """Check if an agent can toggle the object in front of them.

        :param agent_id: The ID of the agent attempting to toggle the object.
        :type agent_id: typing.AgentID
        :return: True if the agent can toggle the object in front of them, False otherwise.
        :rtype: bool
        """
        agent = self.env_agents[agent_id]
        fwd_cell = copy.deepcopy(self.grid.get(*agent.front_pos))
        return fwd_cell.toggle(env=self, toggling_agent=agent)

    def interact(
        self, actions: dict[typing.AgentID, typing.ActionType]
    ) -> None:
        """After agents have moved, let them interact with the environment
        based on their actions (e.g., picking up, dropping, toggling, etc.).

        :param actions: Dictionary of agent IDs and actions.
        :type actions: dict[typing.AgentID, typing.ActionType]
        """
        for a_id, action in actions.items():
            agent: grid_object.GridAgent = self.env_agents[a_id]
            agent.cell_toggled = None
            agent.cell_placed_on = None
            agent.cell_picked_up_from = None
            agent.cell_overlapped = self.grid.get(*agent.pos)

            if action == grid_actions.Actions.RotateRight:
                agent.rotate_right()
            elif action == grid_actions.Actions.RotateLeft:
                agent.rotate_left()

            # Attempt to pick up the object in front of the agent
            elif action == grid_actions.Actions.PickupDrop:
                fwd_pos = agent.front_pos
                fwd_cell = self.grid.get(*fwd_pos)
                agent_ahead = tuple(fwd_pos) in self.agent_pos

                # If there's an agent in front of you, you can't
                # pick up or drop.
                if agent_ahead:
                    continue

                # TODO(chase): we need to fix this logic so that we check if an agent
                # can pick up the type of object that is returned from pick_up_from
                if (
                    fwd_cell
                    and fwd_cell.can_pickup(agent=agent)
                    and agent.can_pickup(grid_object=fwd_cell)
                ):
                    pos = fwd_cell.pos
                    agent.inventory.append(fwd_cell)
                    fwd_cell.pos = None
                    self.grid.set(*pos, None)
                elif (
                    fwd_cell
                    and fwd_cell.can_pickup_from(agent=agent)
                    and agent.can_pickup(grid_object=fwd_cell)
                ):
                    pickup_cell = fwd_cell.pick_up_from(agent=agent)
                    pickup_cell.pos = None
                    agent.inventory.append(pickup_cell)
                    agent.cell_picked_up_from = fwd_cell
                elif not agent_ahead and not fwd_cell and agent.inventory:
                    drop_cell = agent.inventory.pop(0)
                    drop_cell.pos = fwd_pos
                    self.grid.set(fwd_pos[0], fwd_pos[1], drop_cell)
                elif (
                    fwd_cell
                    and agent.inventory
                    and fwd_cell.can_place_on(
                        cell=agent.inventory[0], agent=agent
                    )
                ):
                    drop_cell = agent.inventory.pop(0)
                    drop_cell.pos = fwd_pos
                    fwd_cell.place_on(cell=drop_cell, agent=agent)
                    agent.cell_placed_on = fwd_cell

                self.on_pickup_drop(a_id)

            # Attempt to toggle the object in front of the agent
            elif action == grid_actions.Actions.Toggle:
                fwd_cell = self.grid.get(*agent.front_pos)
                if fwd_cell:
                    toggle_success = fwd_cell.toggle(env=self, agent=agent)
                    if toggle_success:
                        agent.cell_toggled = fwd_cell

                self.on_toggle(a_id)

        self.on_interact(actions)

    def on_interact(
        self, actions: dict[typing.AgentID, typing.ActionType]
    ) -> None:
        """Hook for subclasses to implement custom logic after agents interact with the environment.

        :param actions: Dictionary of agent IDs and actions.
        :type actions: dict[typing.AgentID, typing.ActionType]
        """
        pass

    def on_toggle(self, agent_id: typing.AgentID) -> None:
        """Hook for subclasses to implement custom logic after an agent toggles an object.

        :param agent_id: The ID of the agent toggling the object.
        :type agent_id: typing.AgentID
        """
        pass

    def on_pickup_drop(self, agent_id: typing.AgentID) -> None:
        """Hook for subclasses to implement custom logic after an agent picks up or drops an object.

        :param agent_id: The ID of the agent picking up or dropping an object.
        :type agent_id: typing.AgentID
        """
        pass

    def on_reset(self) -> None:
        """Hook for subclasses to implement custom logic after the environment is reset."""
        pass

    def on_step(self) -> None:
        """Hook for subclasses to implement custom logic after each step."""
        pass

    def on_move(self, agent_id: typing.AgentID) -> None:
        """Hook for subclasses to implement custom logic after an agent moves.

        :param agent_id: The ID of the agent moving.
        :type agent_id: typing.AgentID
        """
        pass

    def get_obs(self) -> dict[typing.AgentID, typing.ObsType]:
        """Fetch new observations for the agents

        :return: Dictionary of agent IDs and their observations.
        :rtype: dict[typing.AgentID, typing.ObsType]
        """
        obs = {}
        for a_id in self.agent_ids:
            obs[a_id] = self.feature_spaces[a_id].generate_features()
        return obs

    def compute_rewards(
        self,
    ) -> None:
        """Compute the per agent and per component rewards for the current state transition
        using the reward modules provided in the environment configuration.

        The rewards are added to self.per_agent_rewards and self.per_component_rewards.
        """

        for reward in self.rewards:
            calculated_rewards = reward.calculate_reward(
                state=self.prev_grid,
                agent_actions=self.prev_actions,
                new_state=self.grid,
            )

            # Add component rewards to per agent reward
            for agent_id, reward_value in calculated_rewards.items():
                self.per_agent_reward[agent_id] += reward_value

            # Save reward by component
            self.per_component_reward[reward.name] = calculated_rewards

        for agent_id, val in self.reward_this_step.items():
            self.per_agent_reward[agent_id] += val

        self.reward_this_step = self.get_empty_reward_dict()

    def get_terminateds_truncateds(
        self,
    ) -> tuple[dict[typing.AgentID, bool], dict[typing.AgentID, bool]]:
        """Determine the done status for each agent."""
        terminateds = {
            agent_id: agent.terminated
            for agent_id, agent in self.env_agents.items()
        }
        # terminateds["__all__"] = all([*terminateds.values()])

        if self.t >= self.max_steps:
            truncateds = {agent_id: True for agent_id in self.agent_ids}
        else:
            truncateds = {agent_id: False for agent_id in self.agent_ids}

        # truncateds["__all__"] = all([*truncateds.values()])

        # Update active agents
        self.agents = [
            agent_id
            for agent_id in self.possible_agents
            if not terminateds[agent_id] and not truncateds[agent_id]
        ]

        return terminateds, truncateds

    def get_infos(
        self, **kwargs
    ) -> dict[typing.AgentID, dict[typing.Any, typing.Any]]:
        """Get info dictionaries for each agent.

        :return: Dictionary keyed by agent IDs containing info dictionaries.
        :rtype: dict[typing.Any, typing.Any]
        """
        infos = {agent_id: {} for agent_id in self._agent_ids}
        for info_key, info_dict in kwargs.items():
            for agent_id, val in info_dict.items():
                assert (
                    agent_id in self._agent_ids
                ), f"Must pass dicts keyed by AgentIDs to get_infos(), got invalid key: {agent_id}"
                infos[agent_id][info_key] = val
        return infos

    def get_empty_reward_dict(self) -> dict[typing.AgentID, float]:
        """Get a dictionary of rewards for each agent, initialized to 0.

        :return: Dictionary of rewards for each agent, initialized to 0.
        :rtype: dict[typing.AgentID, float]
        """
        return {a_id: 0 for a_id in self.agent_ids}

    @property
    def map_with_agents(self) -> np.ndarray:
        """retrieve a version of the environment where 'P' chars have agent IDs.

        :return: Map of the environment with agent IDs.
        :rtype: np.ndarray
        """

        grid_encoding = self.grid.encode(encode_char=True)
        grid = grid_encoding[:, :, 0]

        for a_id, agent in self.env_agents.items():
            if (
                agent is not None
            ):  # will be None before being set by subclassed env
                grid[agent.pos[0], agent.pos[1]] = self.id_to_numeric(a_id)

        return grid

    def select_spawn_point(self) -> tuple[int, int]:
        """Select a spawn point for an agent.

        :return: A spawn point for an agent.
        :rtype: tuple[int, int]
        """
        if self.spawn_points:
            return self.spawn_points.pop(0)

        available_spawns = self.available_positions
        return self.np_random.choice(available_spawns)

    @property
    def available_positions(self) -> list[tuple[int, int]]:
        """Get a list of available positions for agents to spawn.

        :return: List of available positions for agents to spawn.
        :rtype: list[tuple[int, int]]
        """
        spawns = []
        for r in range(self.grid.height):
            for c in range(self.grid.width):
                cell = self.grid.get(row=r, col=c)
                if (r, c) not in self.agent_pos and (
                    not cell or cell.object_id == "floor"
                ):
                    spawns.append((r, c))
        return spawns

    def put_obj(self, obj: grid_object.GridObj, row: int, col: int):
        """Place an object at a specific point in the grid.

        :param obj: The object to place.
        :type obj: grid_object.GridObj
        :param row: The row to place the object.
        :type row: int
        :param col: The column to place the object.
        :type col: int
        """
        self.grid.set(row=row, col=col, v=obj)
        obj.pos = (row, col)
        obj.init_pos = (row, col)

    def get_view_exts(
        self, agent_id: typing.AgentID, agent_view_size: int = None
    ) -> tuple[int, int, int, int]:
        """Get the extents of the square set of tiles visible to the agent
        Note: the bottom extent indices are not included in the set
        if agent_view_size is None, use self.agent_view_size


        :param agent_id: Agent ID of the agent to get the view extents for.
        :type agent_id: typing.AgentID
        :param agent_view_size: View distance, defaults to None
        :type agent_view_size: int, optional
        :raises ValueError: Invalid agent direction.
        :return: Tuple of the top-left and bottom-right extents of the agent's view.
        :rtype: tuple[int, int, int, int]
        """
        agent = self.env_agents[agent_id]
        agent_view_size = agent_view_size or self.agent_view_size

        if agent.dir == directions.Directions.Right:
            topY = agent.pos[0] - agent_view_size // 2
            topX = agent.pos[1]
        elif agent.dir == directions.Directions.Down:
            topY = agent.pos[0]
            topX = agent.pos[1] - agent_view_size // 2
        elif agent.dir == directions.Directions.Left:
            topY = agent.pos[0] - agent_view_size // 2
            topX = agent.pos[1] - agent_view_size + 1
        elif agent.dir == directions.Directions.Up:
            topY = agent.pos[0] - agent_view_size + 1
            topX = agent.pos[1] - agent_view_size // 2
        else:
            raise ValueError("Invalid agent direction.")

        botX = topX + agent_view_size
        botY = topY + agent_view_size

        return topX, topY, botX, botY

    def gen_obs_grid(
        self, agent_id: typing.AgentID, agent_view_size: int = None
    ) -> grid.Grid:
        """Generate the sub-grid observed by a specific agent

        :param agent_id: Agent ID of the agent to generate the observation for.
        :type agent_id: typing.AgentID
        :param agent_view_size: Size of the agents view area, defaults to None
        :type agent_view_size: int, optional
        :return: A sub-grid observed by the agent.
        :rtype: grid.Grid
        """
        topX, topY, *_ = self.get_view_exts(agent_id, agent_view_size)

        agent_view_size = agent_view_size or self.agent_view_size

        grid = self.grid.slice(topX, topY, agent_view_size, agent_view_size)

        assert agent_id in grid.grid_agents

        agent = self.env_agents[agent_id]
        for i in range(agent.dir + 1):
            grid = grid.rotate_left()

        if not self.metadata.get("see_through_walls", True):
            # Mask view from the agents position at the bottom-center of the grid view
            vis_mask = grid.process_vis(agent_pos=(-1, grid.width // 2))
        else:
            vis_mask = np.ones(shape=(grid.height, grid.width), dtype=bool)

        assert len(grid.grid_agents) >= 1
        assert grid.grid_agents[agent_id].dir == directions.Directions.Up

        # NOTE: In Minigrid, they replace the agent's position with the item they're carrying. We don't do that
        #   here. Rather, we'll provide an additional observation space that represents the item(s) in inventory.

        return grid, vis_mask

    def get_pov_render(
        self,
        agent_id: typing.AgentID,
        tile_size: int = CoreConstants.TilePixels,
    ) -> np.ndarray:
        """Render a specific agent's POV

        :param agent_id: Agent ID of the POV to render.
        :type agent_id: typing.AgentID
        :param tile_size: The pixel height/width of each rendered grid cell, defaults to CoreConstants.TilePixels
        :type tile_size: int, optional
        :return: Render of an agent's POV.
        :rtype: np.ndarray
        """
        grid, vis_mask = self.gen_obs_grid(agent_id)
        img = grid.render(
            tile_size,
            highlight_mask=vis_mask,
        )
        return img

    def get_full_render(
        self, highlight: bool = False, tile_size: int = CoreConstants.TilePixels
    ) -> np.ndarray:
        """Generate a render of the full environment.

        :param highlight: Highlight the visible region for each agent, defaults to False.
        :type highlight: bool
        :param tile_size: The pixel height/width of each rendered grid cell, defaults to CoreConstants.TilePixels
        :type tile_size: int, optional
        :return: Render of the full environment.
        :rtype: np.ndarray
        """
        if highlight:
            highlight_mask = np.zeros(
                shape=(self.grid.height, self.grid.width), dtype=bool
            )
            for a_id, agent in self.env_agents.items():
                # Determine cell visibility for the agent
                _, vis_mask = self.gen_obs_grid(a_id)

                # compute the world coordinates of the bottom-left corner of the agent's view area
                f_vec = agent.dir_vec
                r_vec = agent.right_vec
                top_left = (
                    agent.pos
                    + f_vec * (self.agent_view_size - 1)
                    - r_vec * (self.agent_view_size // 2)
                )

                # identify the cells to highlight as visible in the render
                for vis_row in range(self.agent_view_size):
                    for vis_col in range(self.agent_view_size):
                        if not vis_mask[vis_row, vis_col]:
                            continue

                        # compute world coordinates of agent view
                        abs_row, abs_col = (
                            top_left - (f_vec * vis_row) + (r_vec * vis_col)
                        )

                        if abs_row < 0 or abs_row >= self.grid.height:
                            continue
                        if abs_col < 0 or abs_col >= self.grid.width:
                            continue

                        highlight_mask[abs_row, abs_col] = True
        else:
            highlight_mask = None

        img = self.grid.render(
            tile_size=tile_size, highlight_mask=highlight_mask
        )
        return img

    def get_frame(
        self,
        highlight: bool = True,
        tile_size: int = CoreConstants.TilePixels,
        agent_pov: typing.AgentID | None = None,
    ) -> np.ndarray:
        """Return RGB image corresponding to the whole environment or an agent's POV

        :param highlight: Highlight the visible region for each agent, defaults to True
        :type highlight: bool, optional
        :param tile_size: The pixel width height of each grid cell, defaults to CoreConstants.TilePixels
        :type tile_size: int, optional
        :param agent_pov: If specified, gets the frame view for the specified agent, defaults to None
        :type agent_pov: str | None, optional
        :return: The rendered image of the environment or agent perspective.
        :rtype: np.ndarray
        """
        if agent_pov:
            frame = self.get_pov_render(
                agent_id=self.agent_pov, tile_size=tile_size
            )
        else:
            frame = self.get_full_render(
                highlight=highlight, tile_size=tile_size
            )

        return frame

    def render(self) -> None | np.ndarray:
        """Render the environment.

        :return: None if rendering is not enabled, otherwise the rendered image.
        :rtype: None | np.ndarray
        """
        if self.visualizer is not None:
            orientations = {
                self.id_to_numeric(a_id): agent.orientation
                for a_id, agent in self.env_agents.items()
            }
            inventories = {
                self.id_to_numeric(a_id): agent.inventory
                for a_id, agent in self.env_agents.items()
                if len(agent.inventory) > 0
            }
            self.visualizer.render(
                self.map_with_agents,
                self.t,
                orientations=orientations,
                subitems=inventories,
            )

        if self.render_mode is None:
            return

        img = self.get_frame(
            self.metadata.get("highlight", False),
            self.tile_size,
            self.metadata.get("agent_pov", None),
        )

        if self.render_mode == "human":
            if pygame is None:
                raise ImportError(
                    "Must install pygame to use interactive mode."
                )
            # TODO(chase): move all pygame logic to run_interactive.py so it's not needed here.
            if self.render_size is None:
                self.render_size = img.shape[:2]
            if self.window is None:
                pygame.init()
                pygame.display.init()
                self.window = pygame.display.set_mode(
                    (self.screen_size, self.screen_size)
                )
                pygame.display.set_caption(self.name)
            if self.clock is None:
                self.clock = pygame.time.Clock()

            surf = pygame.surfarray.make_surface(img)

            # For some reason, pygame is rotating/flipping the image...
            surf = pygame.transform.flip(surf, False, True)
            surf = pygame.transform.rotate(surf, 270)

            # Create background with mission description
            offset = surf.get_size()[0] * 0.1
            bg = pygame.Surface(
                (
                    int(surf.get_size()[0] + offset),
                    int(surf.get_size()[1] + offset),
                )
            )
            bg.convert()
            bg.fill((255, 255, 255))
            bg.blit(surf, (offset / 2, 0))

            bg = pygame.transform.smoothscale(
                bg, (self.screen_size, self.screen_size)
            )

            font_size = 22
            text = (
                f"Score: {np.round(self.cumulative_score, 2)}"
                + self.render_message
            )

            font = pygame.freetype.SysFont(
                pygame.font.get_default_font(), font_size
            )
            text_rect = font.get_rect(text, size=font_size)
            text_rect.center = bg.get_rect().center
            text_rect.y = bg.get_height() - font_size * 1.5
            font.render_to(bg, text_rect, text, size=font_size)

            self.window.blit(bg, (0, 0))
            pygame.event.pump()
            self.clock.tick(self.metadata["render_fps"])
            pygame.display.update()

        elif self.render_mode == "rgb_array":
            return img

    def close(self):
        if self.window:
            pygame.quit()

    def get_action_mask(self, agent_id):
        raise NotImplementedError

    @property
    def agent_ids(self) -> list:
        return list(self.env_agents.keys())

    @property
    def agent_pos(self) -> list:
        return [
            tuple(agent.pos)
            for agent in self.env_agents.values()
            if agent is not None
        ]

    def id_to_numeric(self, agent_id) -> str:
        """Converts agent id to integer, beginning with 1,
        e.g., agent-0 -> 1, agent-1 -> 2, etc.
        """
        agent = self.env_agents[agent_id]
        return str(agent.agent_number)



import collections

from cogrid.feature_space import feature
from cogrid.feature_space import features
from cogrid.feature_space import feature_space
from cogrid.core import grid_utils
from cogrid import cogrid_env
from cogrid.core import grid_object
from cogrid.envs.overcooked import overcooked_grid_objects
import numpy as np


def euclidian_distance(pos_1: tuple[int, int], pos_2: tuple[int, int]) -> int:
    """Calculate the euclidian distance between two points.

    :param pos_1: The first point on the grid.
    :type pos_1: tuple[int, int]
    :param pos_2: The second point on the grid.
    :type pos_2: tuple[int, int]
    :return: The euclidian distance between the two points.
    :rtype: int
    """
    return np.sqrt((pos_1[0] - pos_2[0]) ** 2 + (pos_1[1] - pos_2[1]) ** 2)


class OvercookedCollectedFeatures(feature.Feature):
    """
    A wrapper class to create all overcooked features as a single array.

    For each agent j, calculate:

        - Agent j Direction
        - Agent j Inventory
        - Agent j Adjacent to Counter
        - Agent j Dist to closest {onion, plate, platestack, onionstack, onionsoup, deliveryzone}
        - Agent j Pot Features
            - pot_j_reachable: {0, 1}
            - pot_j_status: onehot of {empty | full | is_cooking | is_ready}
            - pot_j_contents: integer of the number of onions in the pot
            - pot_j_cooking_timer: integer for the number of ts remaining if cooking, 0 if finished, -1 if not cooking
            - pot_j_distance: (dy, dx) from the player's location
            - pot_j_location: (row, column) of the pot on the grid
        - Agent j Distance to other agents j != i
        - Agent j Position

    The observation is the concatenation of all these features for all players.
    """

    def __init__(self, env: cogrid_env.CoGridEnv, **kwargs):

        num_agents = env.config["num_agents"]

        # NOTE: This is hardcoded because we can have a changing number of pots in the environment.
        # We need it to be the max possible number of pots to keep the feature shape constant.
        max_num_pots = 2

        self.agent_features = [
            features.AgentDir(),
            OvercookedInventory(),
            NextToCounter(),
            NextToPot(),
            ClosestObj(focal_object_type=overcooked_grid_objects.Onion, n=4),
            ClosestObj(focal_object_type=overcooked_grid_objects.Plate, n=4),
            ClosestObj(
                focal_object_type=overcooked_grid_objects.PlateStack, n=2
            ),
            ClosestObj(
                focal_object_type=overcooked_grid_objects.OnionStack, n=2
            ),
            ClosestObj(
                focal_object_type=overcooked_grid_objects.OnionSoup, n=4
            ),
            ClosestObj(
                focal_object_type=overcooked_grid_objects.DeliveryZone, n=2
            ),
            ClosestObj(focal_object_type=grid_object.Counter, n=4),
            OrderedPotFeatures(num_pots=max_num_pots),
            DistToOtherPlayers(num_other_players=num_agents - 1),
            features.AgentPosition(),
            features.CanMoveDirection(),
        ]

        self.global_features = [
            LayoutID(),
            EnvironmentLayout(max_shape=(11, 7)),
        ]

        full_shape = num_agents * np.sum(
            [feature.shape for feature in self.agent_features]
        ) + np.sum([feature.shape for feature in self.global_features])

        super().__init__(
            low=-np.inf,
            high=np.inf,
            shape=(full_shape,),
            name="overcooked_features",
            **kwargs,
        )

    def generate(
        self, env: cogrid_env.CoGridEnv, player_id, **kwargs
    ) -> np.ndarray:
        player_encodings = [self.generate_player_encoding(env, player_id)]

        for pid in env.agent_ids:
            if pid == player_id:
                continue
            player_encodings.append(self.generate_player_encoding(env, pid))

        encoding = np.hstack(player_encodings).astype(np.float32)

        global_encoding = self.generate_global_encoding(env, player_id)
        encoding = np.hstack([encoding, global_encoding])

        assert np.array_equal(self.shape, encoding.shape)

        return encoding

    def generate_player_encoding(
        self, env: cogrid_env.CoGridEnv, player_id: str | int
    ) -> np.ndarray:
        encoded_features = []
        for feature in self.agent_features:
            encoded_features.append(feature.generate(env, player_id))

        return np.hstack(encoded_features)

    def generate_global_encoding(
        self, env: cogrid_env.CoGridEnv, player_id: str | int
    ) -> np.ndarray:
        encoded_features = []
        for feature in self.global_features:
            encoded_features.append(feature.generate(env, player_id))

        return np.hstack(encoded_features)


feature_space.register_feature(
    "overcooked_features", OvercookedCollectedFeatures
)


class LayoutID(feature.Feature):
    shape = (5,)

    def __init__(self, **kwargs):
        super().__init__(low=0, high=1, name="layout_id", **kwargs)
        self.overcooked_layouts = [
            "overcooked_cramped_room_v0",
            "overcooked_asymmetric_advantages_v0",
            "overcooked_coordination_ring_v0",
            "overcooked_forced_coordination_v0",
            "overcooked_counter_circuit_v0",
        ]

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.int32)
        encoding[self.overcooked_layouts.index(env.current_layout_id)] = 1
        assert np.array_equal(self.shape, encoding.shape)
        return encoding


class OvercookedInventory(feature.Feature):
    shape = (3,)

    def __init__(self, **kwargs):
        super().__init__(low=0, high=1, name="overcooked_inventory", **kwargs)

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.int32)
        agent = env.grid.grid_agents[player_id]

        if not agent.inventory:
            return encoding

        objs = [
            overcooked_grid_objects.Onion,
            overcooked_grid_objects.OnionSoup,
            overcooked_grid_objects.Plate,
        ]
        encoding[objs.index(type(agent.inventory[0]))] = 1
        assert np.array_equal(self.shape, encoding.shape)
        return encoding


class NextToCounter(feature.Feature):
    """A feature that represents a multi-hot encoding of whether or not there is a counter
    immediately in each of the four cardinal directions.

    For example, let '#' be the counter and '@' be the player. The following situation
        ####
        #  #
        # @#
        ####

    would result in the feature [0, 1, 1, 0], corresponding to having a counter
    immediately to the east and south, but not north and west.
    """

    shape = (4,)

    def __init__(self, **kwargs):
        super().__init__(low=0, high=1, name="next_to_counter", **kwargs)

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros((4,), dtype=np.int32)
        agent = env.grid.grid_agents[player_id]

        for i, (row, col) in enumerate(
            grid_utils.adjacent_positions(*agent.pos)
        ):
            adj_cell = env.grid.get(row, col)
            if isinstance(adj_cell, grid_object.Counter):
                encoding[i] = 1

        assert np.array_equal(self.shape, encoding.shape)
        return encoding


class NextToPot(feature.Feature):
    """A feature that represents a multi-hot encoding of whether or not there is a pot
    in each of the four cardinal directions, for each of the four pot statuses.

    The feature is a concatenation of four vectors of shape (4,), resulting in a shape of (16,).
    The four vectors correspond to:
    1. Empty pot
    2. Pot with less than three onions
    3. Pot that is currently cooking
    4. Full pot (3 onions, ready to serve)

    For example, if there's an empty pot to the north and a cooking pot to the east,
    the feature would be [1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0].
    """

    shape = (16,)

    def __init__(self, **kwargs):
        super().__init__(low=0, high=1, name="next_to_pot", **kwargs)

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.int32)
        agent = env.grid.grid_agents[player_id]

        for i, (row, col) in enumerate(
            grid_utils.adjacent_positions(*agent.pos)
        ):
            adj_cell = env.grid.get(row, col)
            if isinstance(adj_cell, overcooked_grid_objects.Pot):
                if len(adj_cell.objects_in_pot) == 0:
                    encoding[i] = 1
                elif 0 < len(adj_cell.objects_in_pot) < adj_cell.capacity:
                    encoding[i + 4] = 1
                elif adj_cell.is_cooking:
                    encoding[i + 8] = 1
                elif adj_cell.dish_ready:
                    encoding[i + 12] = 1

        assert np.array_equal(self.shape, encoding.shape)
        return encoding


class ClosestObj(feature.Feature):
    """
    This feature calculates (dy, dx) to the closest instance of a specified object.

    # TODO(chase): Use BFS here, right now we're just doing (agent_y - pos_y, agent_x - pos_x)!
    It uses BFS to calculate a path from the player's position to the target, if
    there is no possible path, returns (0, 0).

    n is the number of closest objects to return.

    Note that this does not consider item's that the agent is holding in its inventory.

    """

    def __init__(
        self, focal_object_type: grid_object.GridObj, n: int = 1, **kwargs
    ):
        super().__init__(
            low=-np.inf,
            high=np.inf,
            name=f"closest_{focal_object_type.object_id}_dist",
            shape=(2 * n,),
            **kwargs,
        )
        self.focal_object_type = focal_object_type
        self.n = n

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        agent = env.grid.grid_agents[player_id]
        encoding = np.zeros(self.shape, dtype=np.int32)

        # collect the distances
        distances: list[int] = []
        deltas: list[tuple[int, int]] = []

        for grid_obj in env.grid.grid:
            if grid_obj is None:
                continue
            # Check if the grid obj is what we're looking for
            is_focal_obj = isinstance(
                grid_obj, self.focal_object_type
            ) and not np.array_equal(agent.pos, grid_obj.pos)

            obj_is_placed_on = isinstance(
                grid_obj.obj_placed_on, self.focal_object_type
            )

            if is_focal_obj or obj_is_placed_on:
                delta = np.array(agent.pos) - np.array(grid_obj.pos)
                manhattan_distance = np.sum(np.abs(delta))
                deltas.append(delta)
                distances.append(manhattan_distance)

        # if there were no instances of that object, return (0, 0)
        if not distances:
            return np.zeros(self.shape, dtype=np.int32)

        # sort the distances and retrieve the corresponding deltas
        sorted_indices = np.argsort(distances)
        sorted_deltas = [deltas[i] for i in sorted_indices][: self.n]

        # flatten the sorted deltas into a flat array
        flat_deltas = [item for sublist in sorted_deltas for item in sublist]

        encoding[: len(flat_deltas)] = flat_deltas

        # # find the closest instance and return that array
        # min_dist_idx = np.argmin(euc_distances)
        # encoding = np.asarray(distances[min_dist_idx], dtype=np.int32)
        assert np.array_equal(self.shape, encoding.shape)

        return encoding


class OrderedPotFeatures(feature.Feature):
    """Encode features related to the pot. Note that this assumes the number of pots is fixed,
    otherwise the feature size will vary and will cause errors. For each pot, calculate:
        - pot_j_reachable: {0, 1}  # TODO(chase): use BFS to calculate this, currently fixed at 1.
        - pot_j_status: onehot of {empty | full | is_cooking | is_ready}
        - pot_j_contents: integer of the number of onions in the pot
        - pot_j_cooking_timer: integer for the number of ts remaining if cooking, 0 if finished, -1 if not cooking
        - pot_j_distance: (dy, dx) from the player's location
        - pot_j_location: (row, column) of the pot on the grid

    We will sort based on the euclidian distance to the player's location and then concatenate all pot features.
    """

    def __init__(self, num_pots=1, **kwargs):
        super().__init__(
            low=-np.inf,
            high=np.inf,
            shape=(num_pots * 11,),
            name="pot_features",
            **kwargs,
        )

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        pot_feature_dict = {}
        agent = env.grid.grid_agents[player_id]
        count = 0
        pot_features = np.zeros(self.shape, dtype=np.float32)
        for grid_obj in env.grid.grid:
            if not isinstance(grid_obj, overcooked_grid_objects.Pot):
                continue
            count += 1

            # Encode if the pot is reachable (size 1)
            pot_reachable = [1]  # TODO(chase): use search to determine

            # Encode if the pot is empty, cooking, or ready (size 4)
            pot_status = np.zeros(
                (4,), dtype=np.int32
            )  # empty, cooking, ready, ptr
            if grid_obj.dish_ready:
                pot_status[0] = 1
            elif len(grid_obj.objects_in_pot) == 0:
                pot_status[1] = 1
            elif len(grid_obj.objects_in_pot) == grid_obj.capacity:
                pot_status[2] = 1
            else:
                pot_status[3] = 1

            # Encode the number of each legal content in the pot (size legal_contents)
            pot_contents = np.zeros((len(grid_obj.legal_contents),))
            item_types_in_pot = [
                grid_obj.legal_contents.index(type(pot_content_obj))
                for pot_content_obj in grid_obj.objects_in_pot
            ]
            for obj_index, obj_count in collections.Counter(
                item_types_in_pot
            ).items():
                pot_contents[obj_index] = obj_count

            # Encode cooking time (size 1)
            pot_cooking_time = np.array(
                (grid_obj.cooking_timer if grid_obj.is_cooking else -1,),
                dtype=np.int32,
            )

            # encode the distance from agent to pot (size 2)
            pot_distance = np.asarray(agent.pos) - np.asarray(grid_obj.pos)

            # encode the pot location (size 2)
            pot_location = np.asarray(grid_obj.pos)

            # concatenate all features and store distance
            euc_dist = euclidian_distance(agent.pos, grid_obj.pos)

            pot_features = np.hstack(
                [
                    pot_reachable,
                    pot_status,
                    pot_contents,
                    pot_cooking_time,
                    pot_distance,
                    pot_location,
                ]
            )
            pot_feature_dict[grid_obj.uuid] = (euc_dist, pot_features)

        # sort based on euclidian distance
        # 1 indexes the value, 0 the euc_distance from above
        # pot_feature_dict = dict(
        #     sorted(pot_feature_dict.items(), key=lambda item: item[1][0])
        # )

        pot_feature_values = [v[1] for v in pot_feature_dict.values()]
        encoding = np.hstack(pot_feature_values)

        padded_encoding = np.zeros(self.shape, dtype=np.float32)
        padded_encoding[: len(encoding)] = encoding

        return padded_encoding


class DistToOtherPlayers(feature.Feature):
    """Return an encoding of the distance to all other players, unsorted."""

    def __init__(self, num_other_agents=1, **kwargs):
        super().__init__(
            low=1,
            high=np.inf,
            shape=(num_other_agents * 2,),
            name="dist_to_other_players",
            **kwargs,
        )
        self.num_other_agents = num_other_agents

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros((2 * (self.num_other_agents),), dtype=np.int32)
        agent = env.grid.grid_agents[player_id]

        other_agent_nums = 0
        for pid, other_agent in env.grid.grid_agents.items():
            if pid == player_id:
                continue

            encoding[other_agent_nums * 2 : other_agent_nums * 2 + 2] = (
                np.asarray(agent.pos) - np.asarray(other_agent.pos)
            )
        other_agent_nums += 1

        assert np.array_equal(self.shape, encoding.shape), (
            "DistToOtherPlayers shape is off. You likely are using an environment with > 2 "
            "agents, in which case you should change the feature encoding."
        )
        return encoding


class EnvironmentLayout(feature.Feature):
    """
    A 1D array representation of the entire environment for each kind of object.

    For each object type, there is a flattened 1D array where each element corresponds
    to a cell in the grid. The value is 1 if the corresponding cell contains that object,
    and 0 otherwise.
    """

    def __init__(self, max_shape, **kwargs):
        self.max_shape = max_shape
        self.object_types = [
            grid_object.Counter,
            overcooked_grid_objects.Pot,
            overcooked_grid_objects.Onion,
            overcooked_grid_objects.Plate,
            overcooked_grid_objects.OnionStack,
            overcooked_grid_objects.PlateStack,
        ]
        shape = (len(self.object_types) * (max_shape[0] * max_shape[1]),)
        super().__init__(
            low=0, high=1, shape=shape, name="environment_layout", **kwargs
        )

    def generate(self, env: cogrid_env.CoGridEnv, player_id, **kwargs):
        encoding = np.zeros(self.shape, dtype=np.int32)
        current_height, current_width = env.grid.height, env.grid.width

        for obj_index, obj_type in enumerate(self.object_types):
            start_index = obj_index * (self.max_shape[0] * self.max_shape[1])
            for row in range(current_height):
                for col in range(current_width):
                    cell = env.grid.get(row, col)
                    if isinstance(cell, obj_type):
                        flat_index = row * self.max_shape[1] + col
                        encoding[start_index + flat_index] = 1

        assert np.array_equal(self.shape, encoding.shape)
        return encoding



from cogrid.core.agent import Agent


class GoalSeekingAgent(Agent):
    def __init__(
        self,
        agent_id,
        start_position,
        env,
        target_values,
        config,
    ):
        super().__init__(agent_id, start_position, env, config)

        self.target_values = target_values

        self.step_penalty = 0.01
        self.collision_penalty = 0.05

    def interact(self, char):
        raise NotImplementedError("Must add in reward modules!")
        # return " "

    def create_inventory_ob(self):
        return [
            1 if obj in self.inventory else 0
            for obj in self.target_values.keys()
        ]

    @property
    def inventory_capacity(self):
        return len([*self.target_values.keys()])



from cogrid.core.agent import Agent
from cogrid.core.grid_object import GridObj
from cogrid.envs.overcooked import overcooked_grid_objects


class OvercookedAgent(Agent):
    def can_pickup(self, grid_object: GridObj) -> bool:
        """Determine if the agent can pickup a specified grid object.
        The Overcooked agent can pick up objects until it reaches capacity,


        :param grid_object: The grid object to check if the agent can pick up.
        :type grid_object: GridObj
        :return: True if the agent can pick up the object, False otherwise.
        :rtype: bool
        """
        if isinstance(grid_object, overcooked_grid_objects.Pot) and any(
            [
                isinstance(inv_obj, overcooked_grid_objects.Plate)
                for inv_obj in self.inventory
            ]
        ):
            return True

        return len(self.inventory) < self.inventory_capacity
